# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3
from pyelasticsearch import ElasticSearch

from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.db_settings import elasticsearch_settings
from testing_support.validators.validate_database_duration import validate_database_duration

from newrelic.api.background_task import background_task

ES_SETTINGS = elasticsearch_settings()[0]
ES_URL = 'http://%s:%s' % (ES_SETTINGS['host'], ES_SETTINGS['port'])

def _exercise_es(es):
    es.index("contacts", "person",
            {"name": "Joe Tester", "age": 25, "title": "QA Engineer"}, id=1)
    es.index("contacts", "person",
            {"name": "Jessica Coder", "age": 32, "title": "Programmer"}, id=2)
    es.index("contacts", "person",
            {"name": "Freddy Tester", "age": 29, "title": "Assistant"}, id=3)
    es.refresh('contacts')
    es.index("address", "employee", {"name": "Sherlock",
        "address": "221B Baker Street, London"}, id=1)
    es.index("address", "employee", {"name": "Bilbo",
        "address": "Bag End, Bagshot row, Hobbiton, Shire"}, id=2)
    es.search('name:Joe', index='contacts')
    es.search('name:jessica', index='contacts')
    es.search('name:Sherlock', index='address')
    es.search('name:Bilbo', index=['contacts', 'address'])
    es.search('name:Bilbo', index='contacts,address')
    es.search('name:Bilbo', index='*')
    es.search('name:Bilbo')
    es.status()

# Common Metrics for tests that use _exercise_es().

_test_pyelasticsearch_scoped_metrics = [
    ('Datastore/statement/Elasticsearch/contacts/index', 3),
    ('Datastore/statement/Elasticsearch/contacts/search', 2),
    ('Datastore/statement/Elasticsearch/address/index', 2),
    ('Datastore/statement/Elasticsearch/address/search', 1),
    ('Datastore/statement/Elasticsearch/_all/search', 2),
    ('Datastore/statement/Elasticsearch/other/search', 2),
    ('Datastore/statement/Elasticsearch/contacts/refresh', 1),
    ('Datastore/statement/Elasticsearch/_all/status', 1),
]

_test_pyelasticsearch_rollup_metrics = [
    ('Datastore/all', 14),
    ('Datastore/allOther', 14),
    ('Datastore/Elasticsearch/all', 14),
    ('Datastore/Elasticsearch/allOther', 14),
    ('Datastore/operation/Elasticsearch/index', 5),
    ('Datastore/operation/Elasticsearch/search', 7),
    ('Datastore/operation/Elasticsearch/refresh', 1),
    ('Datastore/operation/Elasticsearch/status', 1),
    ('Datastore/statement/Elasticsearch/contacts/index', 3),
    ('Datastore/statement/Elasticsearch/contacts/search', 2),
    ('Datastore/statement/Elasticsearch/address/index', 2),
    ('Datastore/statement/Elasticsearch/address/search', 1),
    ('Datastore/statement/Elasticsearch/_all/search', 2),
    ('Datastore/statement/Elasticsearch/other/search', 2),
    ('Datastore/statement/Elasticsearch/contacts/refresh', 1),
    ('Datastore/statement/Elasticsearch/_all/status', 1),
]

@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
        'test_pyelasticsearch:test_pyelasticsearch_operation',
        scoped_metrics=_test_pyelasticsearch_scoped_metrics,
        rollup_metrics=_test_pyelasticsearch_rollup_metrics,
        background_task=True)
@background_task()
def test_pyelasticsearch_operation():
    client = ElasticSearch(ES_URL)
    _exercise_es(client)

@validate_database_duration()
@background_task()
def test_elasticsearch_database_duration():
    client = ElasticSearch(ES_URL)
    _exercise_es(client)

@validate_database_duration()
@background_task()
def test_elasticsearch_and_sqlite_database_duration():

    # Make ElasticSearch queries

    client = ElasticSearch(ES_URL)
    _exercise_es(client)

    # Make sqlite queries

    conn = sqlite3.connect(":memory:")
    cur = conn.cursor()

    cur.execute("CREATE TABLE contacts (name text, age int)")
    cur.execute("INSERT INTO contacts VALUES ('Bob', 22)")

    conn.commit()
    conn.close()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest

from newrelic.api.lambda_handler import lambda_handler
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_event_attributes import validate_transaction_event_attributes

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE_DIR = os.path.normpath(os.path.join(CURRENT_DIR, 'fixtures'))
FIXTURE = os.path.join(FIXTURE_DIR, 'lambda_event_source.json')
tests = {}
events = {}


def _load_tests():
    with open(FIXTURE, 'r') as fh:
        for test in json.loads(fh.read()):
            test_name = test.pop('name')

            test_file = test_name + '.json'
            path = os.path.join(FIXTURE_DIR, 'lambda_event_source', test_file)
            with open(path, 'r') as fh:
                events[test_name] = json.loads(fh.read())

            tests[test_name] = test
    return tests.keys()


class Context(object):
    aws_request_id = 'cookies'
    invoked_function_arn = 'arn'
    function_name = 'cats'
    function_version = '$LATEST'
    memory_limit_in_mb = 128


@lambda_handler()
def handler(event, context):
    return {
        'statusCode': '200',
        'body': '{}',
        'headers': {
            'Content-Type': 'application/json',
            'Content-Length': 2,
        },
    }


@pytest.mark.parametrize('test_name', _load_tests())
def test_lambda_event_source(test_name):
    _exact = {'user': {}, 'intrinsic': {}, 'agent': {}}

    expected_arn = tests[test_name].get('aws.lambda.eventSource.arn', None)
    if expected_arn:
        _exact['agent']['aws.lambda.eventSource.arn'] = expected_arn
    else:
        pytest.skip("Nothing to test!")
        return

    @override_application_settings({'attributes.include': ['aws.*']})
    @validate_transaction_event_attributes({}, exact_attrs=_exact)
    def _test():
        handler(events[test_name], Context)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest
from testing_support.fixtures import (
    dt_enabled,
    override_application_settings,
    validate_error_event_attributes,
)
from testing_support.sample_applications import fully_featured_app
from testing_support.validators.validate_transaction_error_trace_attributes import (
    validate_transaction_error_trace_attributes,
)
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)

WSGI_ATTRIBUTES = [
    "wsgi.input.seconds",
    "wsgi.input.bytes",
    "wsgi.input.calls.read",
    "wsgi.input.calls.readline",
    "wsgi.input.calls.readlines",
    "wsgi.output.seconds",
    "wsgi.output.bytes",
    "wsgi.output.calls.write",
    "wsgi.output.calls.yield",
]
required_attributes = {"agent": WSGI_ATTRIBUTES, "intrinsic": {}, "user": {}}

app = webtest.TestApp(fully_featured_app)


@validate_transaction_event_attributes(required_attributes)
@validate_error_event_attributes(required_attributes)
@validate_transaction_error_trace_attributes(required_attributes)
@override_application_settings({"attributes.include": ["*"]})
@dt_enabled
def test_wsgi_attributes():
    app.post_json("/", {"foo": "bar"}, extra_environ={"n_errors": "1", "err_message": "oops"})



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mysql.connector

from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_database_trace_inputs import validate_database_trace_inputs

from testing_support.db_settings import mysql_settings
from newrelic.api.background_task import background_task

DB_SETTINGS = mysql_settings()
DB_SETTINGS = DB_SETTINGS[0]
DB_NAMESPACE = DB_SETTINGS["namespace"]
DB_PROCEDURE = "hello_" + DB_NAMESPACE

mysql_version = tuple(int(x) for x in mysql.connector.__version__.split(".")[:3])
if mysql_version >= (8, 0, 30):
    _connector_metric_name = 'Function/mysql.connector.pooling:connect'
else:
    _connector_metric_name = 'Function/mysql.connector:connect'

_test_execute_via_cursor_scoped_metrics = [
        (_connector_metric_name, 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/select' % DB_NAMESPACE, 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/insert' % DB_NAMESPACE, 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/update' % DB_NAMESPACE, 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/delete' % DB_NAMESPACE, 1),
        ('Datastore/operation/MySQL/drop', 2),
        ('Datastore/operation/MySQL/create', 2),
        ('Datastore/statement/MySQL/%s/call' % DB_PROCEDURE, 1),
        ('Datastore/operation/MySQL/commit', 2),
        ('Datastore/operation/MySQL/rollback', 1)]

_test_execute_via_cursor_rollup_metrics = [
        ('Datastore/all', 13),
        ('Datastore/allOther', 13),
        ('Datastore/MySQL/all', 13),
        ('Datastore/MySQL/allOther', 13),
        ('Datastore/operation/MySQL/select', 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/select' % DB_NAMESPACE, 1),
        ('Datastore/operation/MySQL/insert', 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/insert' % DB_NAMESPACE, 1),
        ('Datastore/operation/MySQL/update', 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/update' % DB_NAMESPACE, 1),
        ('Datastore/operation/MySQL/delete', 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/delete' % DB_NAMESPACE, 1),
        ('Datastore/statement/MySQL/%s/call' % DB_PROCEDURE, 1),
        ('Datastore/operation/MySQL/call', 1),
        ('Datastore/operation/MySQL/drop', 2),
        ('Datastore/operation/MySQL/create', 2),
        ('Datastore/operation/MySQL/commit', 2),
        ('Datastore/operation/MySQL/rollback', 1)]

@validate_transaction_metrics('test_database:test_execute_via_cursor',
        scoped_metrics=_test_execute_via_cursor_scoped_metrics,
        rollup_metrics=_test_execute_via_cursor_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=dict)
@background_task()
def test_execute_via_cursor(table_name):

    connection = mysql.connector.connect(db=DB_SETTINGS['name'],
            user=DB_SETTINGS['user'], passwd=DB_SETTINGS['password'],
            host=DB_SETTINGS['host'], port=DB_SETTINGS['port'])

    cursor = connection.cursor()

    cursor.execute("""drop table if exists `%s`""" % table_name)

    cursor.execute("""create table %s """
            """(a integer, b real, c text)""" % table_name)

    cursor.executemany("""insert into `%s` """ % table_name +
            """values (%(a)s, %(b)s, %(c)s)""", [dict(a=1, b=1.0, c='1.0'),
            dict(a=2, b=2.2, c='2.2'), dict(a=3, b=3.3, c='3.3')])

    cursor.execute("""select * from %s""" % table_name)

    for row in cursor: pass

    cursor.execute("""update `%s` """ % table_name +
            """set a=%(a)s, b=%(b)s, c=%(c)s where a=%(old_a)s""",
            dict(a=4, b=4.0, c='4.0', old_a=1))

    cursor.execute("""delete from `%s` where a=2""" % table_name)

    cursor.execute("""drop procedure if exists %s""" % DB_PROCEDURE)
    cursor.execute("""CREATE PROCEDURE %s()
                      BEGIN
                        SELECT 'Hello World!';
                      END""" % DB_PROCEDURE)

    cursor.callproc("%s" % DB_PROCEDURE)

    connection.commit()
    connection.rollback()
    connection.commit()

_test_connect_using_alias_scoped_metrics = [
        (_connector_metric_name, 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/select' % DB_NAMESPACE, 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/insert' % DB_NAMESPACE, 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/update' % DB_NAMESPACE, 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/delete' % DB_NAMESPACE, 1),
        ('Datastore/operation/MySQL/drop', 2),
        ('Datastore/operation/MySQL/create', 2),
        ('Datastore/statement/MySQL/%s/call' % DB_PROCEDURE, 1),
        ('Datastore/operation/MySQL/commit', 2),
        ('Datastore/operation/MySQL/rollback', 1)]

_test_connect_using_alias_rollup_metrics = [
        ('Datastore/all', 13),
        ('Datastore/allOther', 13),
        ('Datastore/MySQL/all', 13),
        ('Datastore/MySQL/allOther', 13),
        ('Datastore/operation/MySQL/select', 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/select' % DB_NAMESPACE, 1),
        ('Datastore/operation/MySQL/insert', 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/insert' % DB_NAMESPACE, 1),
        ('Datastore/operation/MySQL/update', 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/update' % DB_NAMESPACE, 1),
        ('Datastore/operation/MySQL/delete', 1),
        ('Datastore/statement/MySQL/datastore_mysql_%s/delete' % DB_NAMESPACE, 1),
        ('Datastore/statement/MySQL/%s/call' % DB_PROCEDURE, 1),
        ('Datastore/operation/MySQL/call', 1),
        ('Datastore/operation/MySQL/drop', 2),
        ('Datastore/operation/MySQL/create', 2),
        ('Datastore/operation/MySQL/commit', 2),
        ('Datastore/operation/MySQL/rollback', 1)]

@validate_transaction_metrics('test_database:test_connect_using_alias',
        scoped_metrics=_test_connect_using_alias_scoped_metrics,
        rollup_metrics=_test_connect_using_alias_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=dict)
@background_task()
def test_connect_using_alias(table_name):

    connection = mysql.connector.connect(db=DB_SETTINGS['name'],
            user=DB_SETTINGS['user'], passwd=DB_SETTINGS['password'],
            host=DB_SETTINGS['host'], port=DB_SETTINGS['port'])

    cursor = connection.cursor()

    cursor.execute("""drop table if exists `%s`""" % table_name)

    cursor.execute("""create table %s """
            """(a integer, b real, c text)""" % table_name)

    cursor.executemany("""insert into `%s` """ % table_name +
            """values (%(a)s, %(b)s, %(c)s)""", [dict(a=1, b=1.0, c='1.0'),
            dict(a=2, b=2.2, c='2.2'), dict(a=3, b=3.3, c='3.3')])

    cursor.execute("""select * from %s""" % table_name)

    for row in cursor: pass

    cursor.execute("""update `%s` """ % table_name +
            """set a=%(a)s, b=%(b)s, c=%(c)s where a=%(old_a)s""",
            dict(a=4, b=4.0, c='4.0', old_a=1))

    cursor.execute("""delete from `%s` where a=2""" % table_name)

    cursor.execute("""drop procedure if exists %s""" % DB_PROCEDURE)
    cursor.execute("""CREATE PROCEDURE %s()
                      BEGIN
                        SELECT 'Hello World!';
                      END""" % DB_PROCEDURE)

    cursor.callproc("%s" % DB_PROCEDURE)

    connection.commit()
    connection.rollback()
    connection.commit()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.asgi_application import ASGIApplicationWrapper
from newrelic.common.package_version_utils import get_package_version


@property
def application(self):
    return getattr(self, "_nr_application", vars(self).get("application", None))


@application.setter
def application(self, value):
    dispatcher_details = ("Daphne", get_package_version("daphne"))
    # Wrap app only once
    if value and not getattr(value, "_nr_wrapped", False):
        value = ASGIApplicationWrapper(value, dispatcher=dispatcher_details)
        value._nr_wrapped = True
    self._nr_application = value


def instrument_daphne_server(module):
    module.Server.application = application



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import os
import time


from testing_support.fixtures import collector_agent_registration_fixture
from newrelic.core.agent_protocol import AgentProtocol
from newrelic.common.agent_http import HttpClient
from newrelic.core.config import global_settings


class FullUriClient(HttpClient):
    def send_request(
        self, method="POST", path="/agent_listener/invoke_raw_method", *args, **kwargs
    ):
        path = "https://" + self._host + path
        return super(FullUriClient, self).send_request(method, path, *args, **kwargs)


_default_settings = {
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
    "startup_timeout": 10.0,
}
application = collector_agent_registration_fixture(
    app_name="Python Agent Test (test_full_uri_payloads)",
    default_settings=_default_settings,
)


@pytest.fixture
def session(application):
    session = application._agent.application(application.name)._active_session
    return session


NOW = time.time()
EMPTY_SAMPLES = {
    "reservoir_size": 100,
    "events_seen": 0,
}


@pytest.mark.skipif(
    "NEW_RELIC_LICENSE_KEY" not in os.environ,
    reason="License key is not expected to be valid",
)
@pytest.mark.parametrize(
    "method,payload",
    [
        ("metric_data", []),
        ("analytic_event_data", []),
        ("custom_event_data", []),
        ("error_event_data", []),
        ("transaction_sample_data", []),
        ("sql_trace_data", []),
        ("get_agent_commands", []),
        ("profile_data", []),
        ("error_data", []),
        ("span_event_data", []),
        ("agent_command_results", ["", {"0": {}}]),
    ],
)
def test_full_uri_payload(session, method, payload):
    redirect_host = session._protocol.client._host
    if method == "agent_command_results":
        payload[0] = session.configuration.agent_run_id
    protocol = AgentProtocol(
        session.configuration, redirect_host, client_cls=FullUriClient
    )
    # An exception will be raised here if there's a problem with the response
    protocol.send(method, payload)


@pytest.mark.skipif(
    "NEW_RELIC_LICENSE_KEY" not in os.environ,
    reason="License key is not expected to be valid",
)
def test_full_uri_connect():
    # An exception will be raised here if there's a problem with the response
    AgentProtocol.connect(
        "Python Agent Test (test_full_uri_payloads)",
        [],
        [],
        global_settings(),
        client_cls=FullUriClient,
    )



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import logging
import os
import subprocess
import sys
import threading
import time

import pytest

try:
    from Queue import Queue
except ImportError:
    from queue import Queue

from testing_support.sample_applications import (
    error_user_params_added,
    user_attributes_added,
)

from newrelic.admin.record_deploy import record_deploy
from newrelic.api.application import (
    application_instance,
    application_settings,
    register_application,
)
from newrelic.common.agent_http import DeveloperModeClient
from newrelic.common.encoding_utils import (
    deobfuscate,
    json_decode,
    json_encode,
    obfuscate,
    unpack_field,
)
from newrelic.common.object_names import callable_name
from newrelic.common.object_wrapper import (
    ObjectProxy,
    function_wrapper,
    transient_function_wrapper,
    wrap_function_wrapper,
)
from newrelic.config import initialize
from newrelic.core.agent import shutdown_agent
from newrelic.core.attribute import create_attributes
from newrelic.core.attribute_filter import (
    DST_ERROR_COLLECTOR,
    DST_TRANSACTION_TRACER,
    AttributeFilter,
)
from newrelic.core.config import apply_config_setting, flatten_settings, global_settings
from newrelic.core.database_utils import SQLConnections

# from newrelic.core.internal_metrics import InternalTraceContext
# from newrelic.core.stats_engine import CustomMetrics
from newrelic.network.exceptions import RetryDataForRequest
from newrelic.packages import six

_logger = logging.getLogger("newrelic.tests")


def _environ_as_bool(name, default=False):
    flag = os.environ.get(name, default)
    if default is None or default:
        try:
            flag = not flag.lower() in ["off", "false", "0"]
        except AttributeError:
            pass
    else:
        try:
            flag = flag.lower() in ["on", "true", "1"]
        except AttributeError:
            pass
    return flag


if _environ_as_bool("NEW_RELIC_HIGH_SECURITY"):
    DeveloperModeClient.RESPONSES["connect"]["high_security"] = True


def initialize_agent(app_name=None, default_settings=None):
    default_settings = default_settings or {}
    settings = global_settings()

    settings.app_name = "Python Agent Test"

    if "NEW_RELIC_LICENSE_KEY" not in os.environ:
        settings.developer_mode = True
        settings.license_key = "DEVELOPERMODELICENSEKEY"

    settings.startup_timeout = float(os.environ.get("NEW_RELIC_STARTUP_TIMEOUT", 20.0))
    settings.shutdown_timeout = float(os.environ.get("NEW_RELIC_SHUTDOWN_TIMEOUT", 20.0))

    # Disable the harvest thread during testing so that harvest is explicitly
    # called on test shutdown
    settings.debug.disable_harvest_until_shutdown = True

    if app_name is not None:
        settings.app_name = app_name

    for name, value in default_settings.items():
        apply_config_setting(settings, name, value)

    env_directory = os.environ.get("TOX_ENVDIR", None)

    if env_directory is not None:
        log_directory = os.path.join(env_directory, "log")
    else:
        log_directory = "."

    log_file = os.path.join(log_directory, "python-agent-test.log")
    if "GITHUB_ACTIONS" in os.environ:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    try:
        os.unlink(log_file)
    except OSError:
        pass

    class FilteredStreamHandler(logging.StreamHandler):
        def emit(self, record):
            if len(logging.root.handlers) != 0:
                return

            if record.name.startswith("newrelic.packages"):
                return

            if record.levelno < logging.WARNING:
                return

            logging.StreamHandler.emit(self, record)

    _stdout_logger = logging.getLogger("newrelic")
    _stdout_handler = FilteredStreamHandler(sys.stderr)
    _stdout_format = "%(levelname)s - %(message)s"
    _stdout_formatter = logging.Formatter(_stdout_format)
    _stdout_handler.setFormatter(_stdout_formatter)
    _stdout_logger.addHandler(_stdout_handler)

    initialize(log_file=log_file, log_level=log_level, ignore_errors=False)


def capture_harvest_errors():
    queue = Queue()

    def wrap_harvest_loop(wrapped, instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        except Exception:
            exc_info = sys.exc_info()
            queue.put(exc_info)
            raise

    def wrap_shutdown_agent(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)
        if not queue.empty():
            exc_info = queue.get()
            raise exc_info[1]
        return result

    def wrap_record_custom_metric(wrapped, instance, args, kwargs):
        def _bind_params(name, value, *args, **kwargs):
            return name

        metric_name = _bind_params(*args, **kwargs)
        if (
            metric_name.startswith("Supportability/Python/Harvest/Exception")
            and not metric_name.endswith("DiscardDataForRequest")
            and not metric_name.endswith("RetryDataForRequest")
            and not metric_name.endswith(("newrelic.packages.urllib3.exceptions:ClosedPoolError"))
        ):
            exc_info = sys.exc_info()
            queue.put(exc_info)

        return wrapped(*args, **kwargs)

    # Capture all unhandled exceptions from the harvest thread

    wrap_function_wrapper("newrelic.core.agent", "Agent._harvest_loop", wrap_harvest_loop)

    # Treat custom exception metrics as unhandled errors

    wrap_function_wrapper("newrelic.core.stats_engine", "CustomMetrics.record_custom_metric", wrap_record_custom_metric)

    # Re-raise exceptions in the main thread

    wrap_function_wrapper("newrelic.core.agent", "Agent.shutdown_agent", wrap_shutdown_agent)


def collector_agent_registration_fixture(
    app_name=None, default_settings=None, linked_applications=None, should_initialize_agent=True
):
    default_settings = default_settings or {}
    linked_applications = linked_applications or []

    @pytest.fixture(scope="session")
    def _collector_agent_registration_fixture(request):
        if should_initialize_agent:
            initialize_agent(app_name=app_name, default_settings=default_settings)

        settings = global_settings()

        # Determine if should be using an internal fake local
        # collector for the test.

        use_fake_collector = _environ_as_bool("NEW_RELIC_FAKE_COLLECTOR", False)
        use_developer_mode = _environ_as_bool("NEW_RELIC_DEVELOPER_MODE", use_fake_collector)

        # Catch exceptions in the harvest thread and reraise them in the main
        # thread. This way the tests will reveal any unhandled exceptions in
        # either of the two agent threads.

        capture_harvest_errors()

        # Associate linked applications.

        application = application_instance()

        for name in linked_applications:
            application.link_to_application(name)

        # Force registration of the application.

        application = register_application()

        # Attempt to record deployment marker for test. It's ok
        # if the deployment marker does not record successfully.

        api_host = settings.host

        if api_host is None:
            api_host = "api.newrelic.com"
        elif api_host == "staging-collector.newrelic.com":
            api_host = "staging-api.newrelic.com"

        if not use_fake_collector and not use_developer_mode:
            description = os.path.basename(os.path.normpath(sys.prefix))
            try:
                _logger.debug("Record deployment marker host: %s", api_host)
                record_deploy(
                    host=api_host,
                    api_key=settings.api_key,
                    app_name=settings.app_name,
                    description=description,
                    port=settings.port or 443,
                    proxy_scheme=settings.proxy_scheme,
                    proxy_host=settings.proxy_host,
                    proxy_user=settings.proxy_user,
                    proxy_pass=settings.proxy_pass,
                    timeout=settings.agent_limits.data_collector_timeout,
                    ca_bundle_path=settings.ca_bundle_path,
                    disable_certificate_validation=settings.debug.disable_certificate_validation,
                )
            except Exception:
                _logger.exception("Unable to record deployment marker.")

        def finalize():
            shutdown_agent()

        request.addfinalizer(finalize)

        return application

    return _collector_agent_registration_fixture


@pytest.fixture(scope="function")
def collector_available_fixture(request):
    application = application_instance()
    active = application.active
    assert active


def raise_background_exceptions(timeout=5.0):
    @function_wrapper
    def _raise_background_exceptions(wrapped, instance, args, kwargs):
        if getattr(raise_background_exceptions, "enabled", None) is None:
            raise_background_exceptions.event = threading.Event()
        else:
            assert raise_background_exceptions.count == 0

        raise_background_exceptions.enabled = True
        raise_background_exceptions.count = 0
        raise_background_exceptions.exception = None
        raise_background_exceptions.event.clear()

        try:
            result = wrapped(*args, **kwargs)

        except:
            # There was an exception in the immediate decorators.
            # Raise it rather than those from background threads.

            raise_background_exceptions.event.clear()
            raise_background_exceptions.exception = None
            raise

        else:
            # Immediate decorators completed normally. We need
            # though to make sure that background threads
            # completed within the timeout period and that no
            # exception occurred in the background threads.

            raise_background_exceptions.enabled = False

            done = raise_background_exceptions.event.is_set()
            raise_background_exceptions.event.clear()

            exc_info = raise_background_exceptions.exception
            raise_background_exceptions.exception = None

            assert done, "Timeout waiting for background task to finish."

            if exc_info is not None:
                six.reraise(*exc_info)

        return result

    return _raise_background_exceptions


def wait_for_background_threads(timeout=5.0):
    @function_wrapper
    def _wait_for_background_threads(wrapped, instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        finally:
            raise_background_exceptions.event.wait(timeout)

    return _wait_for_background_threads


@function_wrapper
def catch_background_exceptions(wrapped, instance, args, kwargs):
    if not getattr(raise_background_exceptions, "enabled", False):
        return wrapped(*args, **kwargs)

    raise_background_exceptions.count += 1

    try:
        return wrapped(*args, **kwargs)
    except:
        raise_background_exceptions.exception = sys.exc_info()
        raise
    finally:
        raise_background_exceptions.count -= 1
        if raise_background_exceptions.count == 0:
            raise_background_exceptions.event.set()


def make_cross_agent_headers(payload, encoding_key, cat_id):
    value = obfuscate(json_encode(payload), encoding_key)
    id_value = obfuscate(cat_id, encoding_key)
    return {"X-NewRelic-Transaction": value, "X-NewRelic-ID": id_value}


def make_synthetics_header(account_id, resource_id, job_id, monitor_id, encoding_key, version=1):
    value = [version, account_id, resource_id, job_id, monitor_id]
    value = obfuscate(json_encode(value), encoding_key)
    return {"X-NewRelic-Synthetics": value}


def capture_transaction_metrics(metrics_list, full_metrics=None):
    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    @catch_background_exceptions
    def _capture_transaction_metrics(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            metrics = instance.stats_table
            if full_metrics is not None:
                full_metrics.update(metrics)
            for metric in metrics.keys():
                metrics_list.append(metric)
            metrics_list.sort()

        return result

    return _capture_transaction_metrics


def check_event_attributes(event_data, required_params=None, forgone_params=None, exact_attrs=None):
    """Check the event attributes from a single (first) event in a
    SampledDataSet. If necessary, clear out previous errors from StatsEngine
    prior to saving error, so that the desired error is the only one present
    in the data set.
    """
    required_params = required_params or {}
    forgone_params = forgone_params or {}
    exact_attrs = exact_attrs or {}

    intrinsics, user_attributes, agent_attributes = next(iter(event_data))

    if required_params:
        for param in required_params["agent"]:
            assert param in agent_attributes, (param, agent_attributes)
        for param in required_params["user"]:
            assert param in user_attributes, (param, user_attributes)
        for param in required_params["intrinsic"]:
            assert param in intrinsics, (param, intrinsics)

    if forgone_params:
        for param in forgone_params["agent"]:
            assert param not in agent_attributes, (param, agent_attributes)
        for param in forgone_params["user"]:
            assert param not in user_attributes, (param, user_attributes)
        for param in forgone_params["intrinsic"]:
            assert param not in intrinsics, (param, intrinsics)

    if exact_attrs:
        for param, value in exact_attrs["agent"].items():
            assert agent_attributes[param] == value, ((param, value), agent_attributes)
        for param, value in exact_attrs["user"].items():
            assert user_attributes[param] == value, ((param, value), user_attributes)
        for param, value in exact_attrs["intrinsic"].items():
            assert intrinsics[param] == value, ((param, value), intrinsics)


def check_error_attributes(
    parameters, required_params=None, forgone_params=None, exact_attrs=None, is_transaction=True
):
    required_params = required_params or {}
    forgone_params = forgone_params or {}
    exact_attrs = exact_attrs or {}

    parameter_fields = ["userAttributes"]
    if is_transaction:
        parameter_fields.extend(["stack_trace", "agentAttributes", "intrinsics"])

    for field in parameter_fields:
        assert field in parameters

    # we can remove this after agent attributes transition is all over
    assert "parameter_groups" not in parameters
    assert "custom_params" not in parameters
    assert "request_params" not in parameters
    assert "request_uri" not in parameters

    check_attributes(parameters, required_params, forgone_params, exact_attrs)


def check_attributes(parameters, required_params=None, forgone_params=None, exact_attrs=None):
    required_params = required_params or {}
    forgone_params = forgone_params or {}
    exact_attrs = exact_attrs or {}

    intrinsics = parameters.get("intrinsics", {})
    user_attributes = parameters.get("userAttributes", {})
    agent_attributes = parameters.get("agentAttributes", {})

    if required_params:
        for param in required_params["agent"]:
            assert param in agent_attributes, (param, agent_attributes)
        for param in required_params["user"]:
            assert param in user_attributes, (param, user_attributes)
        for param in required_params["intrinsic"]:
            assert param in intrinsics, (param, intrinsics)

    if forgone_params:
        for param in forgone_params["agent"]:
            assert param not in agent_attributes, (param, agent_attributes)
        for param in forgone_params["user"]:
            assert param not in user_attributes, (param, user_attributes)
        for param in forgone_params["intrinsic"]:
            assert param not in intrinsics, (param, intrinsics)

    if exact_attrs:
        for param, value in exact_attrs["agent"].items():
            assert agent_attributes[param] == value, ((param, value), agent_attributes)
        for param, value in exact_attrs["user"].items():
            assert user_attributes[param] == value, ((param, value), user_attributes)
        for param, value in exact_attrs["intrinsic"].items():
            assert intrinsics[param] == value, ((param, value), intrinsics)


def validate_custom_event_collector_json(num_events=1):
    """Validate the format, types and number of custom events."""

    @transient_function_wrapper("newrelic.core.application", "Application.record_transaction")
    def _validate_custom_event_collector_json(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            stats = instance._stats_engine
            settings = stats.settings

            agent_run_id = 666
            sampling_info = stats.custom_events.sampling_info
            samples = list(stats.custom_events)

            # Emulate the payload used in data_collector.py

            payload = (agent_run_id, sampling_info, samples)
            collector_json = json_encode(payload)

            decoded_json = json.loads(collector_json)

            decoded_agent_run_id = decoded_json[0]
            decoded_sampling_info = decoded_json[1]
            decoded_events = decoded_json[2]

            assert decoded_agent_run_id == agent_run_id
            assert decoded_sampling_info == sampling_info

            max_setting = settings.event_harvest_config.harvest_limits.custom_event_data
            assert decoded_sampling_info["reservoir_size"] == max_setting

            assert decoded_sampling_info["events_seen"] == num_events
            assert len(decoded_events) == num_events

            for intrinsics, attributes in decoded_events:
                assert isinstance(intrinsics, dict)
                assert isinstance(attributes, dict)

        return result

    return _validate_custom_event_collector_json


def validate_tt_parameters(required_params=None, forgone_params=None):
    required_params = required_params or {}
    forgone_params = forgone_params or {}

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_tt_parameters(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            # Now that transaction has been recorded, generate
            # a transaction trace

            connections = SQLConnections()
            trace_data = instance.transaction_trace_data(connections)
            pack_data = unpack_field(trace_data[0][4])
            tt_intrinsics = pack_data[0][4]["intrinsics"]

            for name in required_params:
                assert name in tt_intrinsics, "name=%r, intrinsics=%r" % (name, tt_intrinsics)
                assert tt_intrinsics[name] == required_params[name], "name=%r, value=%r, intrinsics=%r" % (
                    name,
                    required_params[name],
                    tt_intrinsics,
                )

            for name in forgone_params:
                assert name not in tt_intrinsics, "name=%r, intrinsics=%r" % (name, tt_intrinsics)

        return result

    return _validate_tt_parameters


def validate_tt_segment_params(forgone_params=(), present_params=(), exact_params=None):
    exact_params = exact_params or {}
    recorded_traces = []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _extract_trace(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        # Now that transaction has been recorded, generate
        # a transaction trace

        connections = SQLConnections()
        trace_data = instance.transaction_trace_data(connections)
        # Save the recorded traces
        recorded_traces.extend(trace_data)

        return result

    @function_wrapper
    def validator(wrapped, instance, args, kwargs):
        new_wrapper = _extract_trace(wrapped)
        result = new_wrapper(*args, **kwargs)

        # Verify that traces have been recorded
        assert recorded_traces

        # Extract the first transaction trace
        transaction_trace = recorded_traces[0]
        pack_data = unpack_field(transaction_trace[4])

        # Extract the root segment from the root node
        root_segment = pack_data[0][3]

        recorded_params = {}

        def _validate_segment_params(segment):
            segment_params = segment[3]

            # Translate from the string cache
            for key, value in segment_params.items():
                if hasattr(value, "startswith") and value.startswith("`"):
                    try:
                        index = int(value[1:])
                        value = pack_data[1][index]
                    except ValueError:
                        pass
                segment_params[key] = value

            recorded_params.update(segment_params)

            for child_segment in segment[4]:
                _validate_segment_params(child_segment)

        _validate_segment_params(root_segment)

        recorded_params_set = set(recorded_params.keys())

        # Verify that the params in present params have been recorded
        present_params_set = set(present_params)
        assert recorded_params_set.issuperset(present_params_set)

        # Verify that all forgone params are omitted
        recorded_forgone_params = recorded_params_set & set(forgone_params)
        assert not recorded_forgone_params

        # Verify that all exact params are correct
        for key, value in exact_params.items():
            assert recorded_params[key] == value

        return result

    return validator


def validate_browser_attributes(required_params=None, forgone_params=None):
    required_params = required_params or {}
    forgone_params = forgone_params or {}

    @transient_function_wrapper("newrelic.api.web_transaction", "WSGIWebTransaction.browser_timing_footer")
    def _validate_browser_attributes(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise

        # pick out attributes from footer string_types

        footer_data = result.split("NREUM.info=")[1]
        footer_data = footer_data.split("</script>")[0]
        footer_data = json.loads(footer_data)

        if "intrinsic" in required_params:
            for attr in required_params["intrinsic"]:
                assert attr in footer_data

        if "atts" in footer_data:
            obfuscation_key = instance._settings.license_key[:13]
            attributes = json_decode(deobfuscate(footer_data["atts"], obfuscation_key))
        else:
            # if there are no user or agent attributes, there will be no dict
            # for them in the browser data

            attributes = None

        if "user" in required_params:
            for attr in required_params["user"]:
                assert attr in attributes["u"]

        if "agent" in required_params:
            for attr in required_params["agent"]:
                assert attr in attributes["a"]

        if "user" in forgone_params:
            if attributes:
                if "u" in attributes:
                    for attr in forgone_params["user"]:
                        assert attr not in attributes["u"]

        if "agent" in forgone_params:
            if attributes:
                if "a" in attributes:
                    for attr in forgone_params["agent"]:
                        assert attr not in attributes["a"]

        return result

    return _validate_browser_attributes


def validate_error_event_attributes(required_params=None, forgone_params=None, exact_attrs=None):
    """Check the error event for attributes, expect only one error to be
    present in the transaction.
    """
    required_params = required_params or {}
    forgone_params = forgone_params or {}
    exact_attrs = exact_attrs or {}
    error_data_samples = []

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
        def _validate_error_event_attributes(wrapped, instance, args, kwargs):
            try:
                result = wrapped(*args, **kwargs)
            except:
                raise
            else:
                event_data = instance.error_events
                for sample in event_data:
                    error_data_samples.append(sample)

                check_event_attributes(event_data, required_params, forgone_params, exact_attrs)

            return result

        _new_wrapper = _validate_error_event_attributes(wrapped)
        val = _new_wrapper(*args, **kwargs)
        assert error_data_samples
        return val

    return _validate_wrapper


def validate_error_trace_attributes_outside_transaction(
    err_name, required_params=None, forgone_params=None, exact_attrs=None
):
    required_params = required_params or {}
    forgone_params = forgone_params or {}
    exact_attrs = exact_attrs or {}

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.notice_error")
    def _validate_error_trace_attributes_outside_transaction(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            target_error = core_application_stats_engine_error(err_name)

            check_error_attributes(
                target_error.parameters, required_params, forgone_params, exact_attrs, is_transaction=False
            )

        return result

    return _validate_error_trace_attributes_outside_transaction


def validate_error_event_attributes_outside_transaction(
    required_params=None, forgone_params=None, exact_attrs=None, num_errors=None
):
    required_params = required_params or {}
    forgone_params = forgone_params or {}

    event_data = []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.notice_error")
    def _validate_error_event_attributes_outside_transaction(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            for event in instance.error_events:
                event_data.append(event)

        return result

    @function_wrapper
    def wrapper(wrapped, instance, args, kwargs):
        try:
            result = _validate_error_event_attributes_outside_transaction(wrapped)(*args, **kwargs)
        except:
            raise
        else:
            if num_errors is not None:
                exc_message = (
                    "Expected: %d, Got: %d. Verify StatsEngine is being reset before using this validator."
                    % (num_errors, len(event_data))
                )
                assert num_errors == len(event_data), exc_message

            for event in event_data:
                check_event_attributes([event], required_params, forgone_params, exact_attrs=exact_attrs)

        return result

    return wrapper


def validate_request_params_omitted():
    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_request_params(wrapped, instance, args, kwargs):
        def _bind_params(transaction, *args, **kwargs):
            return transaction

        transaction = _bind_params(*args, **kwargs)

        for attr in transaction.agent_attributes:
            assert not attr.name.startswith("request.parameters")

        return wrapped(*args, **kwargs)

    return _validate_request_params


def validate_attributes(attr_type, required_attr_names=None, forgone_attr_names=None):
    required_attr_names = required_attr_names or []
    forgone_attr_names = forgone_attr_names or []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_attributes(wrapped, instance, args, kwargs):
        def _bind_params(transaction, *args, **kwargs):
            return transaction

        transaction = _bind_params(*args, **kwargs)

        if attr_type == "intrinsic":
            attributes = transaction.trace_intrinsics
            attribute_names = attributes.keys()
        elif attr_type == "agent":
            attributes = transaction.agent_attributes
            attribute_names = [a.name for a in attributes]
            # validate that all agent attributes are included on the RootNode
            root_attribute_names = transaction.root.agent_attributes.keys()
            for name in attribute_names:
                assert name in root_attribute_names, name
        elif attr_type == "user":
            attributes = transaction.user_attributes
            attribute_names = [a.name for a in attributes]

            # validate that all user attributes are included on the RootNode
            root_attribute_names = transaction.root.user_attributes.keys()
            for name in attribute_names:
                assert name in root_attribute_names, name
        for name in required_attr_names:
            assert name in attribute_names, "name=%r, attributes=%r" % (name, attributes)

        for name in forgone_attr_names:
            assert name not in attribute_names, "name=%r, attributes=%r" % (name, attributes)

        return wrapped(*args, **kwargs)

    return _validate_attributes


def validate_attributes_complete(attr_type, required_attrs=None, forgone_attrs=None):
    # This differs from `validate_attributes` in that all fields of
    # Attribute must match (name, value, and destinations), not just
    # name. It's a more thorough test, but it's more of a pain to set
    # up, since you have to pass lists of Attributes to required_attrs
    # and forgone_attrs. For required destinations, the attribute will
    # match if at least the required destinations are present. For
    # forgone attributes the test will fail if any of the destinations
    # provided in the forgone attribute are found.
    #
    # Args:
    #
    #       attr_type: 'intrinsic' or 'agent' or 'user'
    #       required_attrs: List of Attributes that must be present.
    #       forgone_attrs: List of Attributes that must NOT be present.
    #
    # Note:
    #
    # The 'intrinsics' come from `transaction.trace_intrinsics`.

    required_attrs = required_attrs or []
    forgone_attrs = forgone_attrs or []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_attributes_complete(wrapped, instance, args, kwargs):
        def _bind_params(transaction, *args, **kwargs):
            return transaction

        transaction = _bind_params(*args, **kwargs)
        attribute_filter = transaction.settings.attribute_filter

        if attr_type == "intrinsic":
            # Intrinsics are stored as a dict, so for consistency's sake
            # in this test, we convert them to Attributes.

            items = transaction.trace_intrinsics
            attributes = create_attributes(items, DST_ERROR_COLLECTOR | DST_TRANSACTION_TRACER, attribute_filter)

        elif attr_type == "agent":
            attributes = transaction.agent_attributes

        elif attr_type == "user":
            attributes = transaction.user_attributes

        def _find_match(a, attributes):
            # Match by name and value. Ignore destination.
            return next((match for match in attributes if match.name == a.name and match.value == a.value), None)

        # Check that there is a name/value match, and that the destinations
        # for the matched attribute include the ones in required.

        for required in required_attrs:
            match = _find_match(required, attributes)
            assert match, "required=%r, attributes=%r" % (required, attributes)

            result_dest = required.destinations & match.destinations
            assert result_dest == required.destinations, "required=%r, attributes=%r" % (required, attributes)

        # Check that the name and value are NOT going to ANY of the
        # destinations provided as forgone, either because there is no
        # name/value match, or because there is a name/value match, but
        # the destinations do not include the ones in forgone.

        for forgone in forgone_attrs:
            match = _find_match(forgone, attributes)

            if match:
                result_dest = forgone.destinations & match.destinations
                assert result_dest == 0, "forgone=%r, attributes=%r" % (forgone, attributes)

        return wrapped(*args, **kwargs)

    return _validate_attributes_complete


def validate_agent_attribute_types(required_attrs):
    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_agent_attribute_types(wrapped, instance, args, kwargs):
        def _bind_params(transaction, *args, **kwargs):
            return transaction

        transaction = _bind_params(*args, **kwargs)
        attributes = transaction.agent_attributes
        attr_vals = {}
        for attr in attributes:
            attr_vals[attr.name] = attr.value

        for attr_name, attr_type in required_attrs.items():
            assert attr_name in attr_vals
            assert isinstance(attr_vals[attr_name], attr_type)

        return wrapped(*args, **kwargs)

    return _validate_agent_attribute_types


def validate_transaction_event_sample_data(required_attrs, required_user_attrs=True):
    """This test depends on values in the test application from
    agent_features/test_analytics.py, and is only meant to be run as a
    validation with those tests.
    """

    add_transaction_called = []

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        @transient_function_wrapper("newrelic.core.stats_engine", "SampledDataSet.add")
        def _validate_transaction_event_sample_data(wrapped, instance, args, kwargs):
            def _bind_params(sample, *args, **kwargs):
                return sample

            sample = _bind_params(*args, **kwargs)

            assert isinstance(sample, list)
            assert len(sample) == 3

            intrinsics, user_attributes, _ = sample

            if intrinsics["type"] != "Transaction":
                return wrapped(*args, **kwargs)  # Exit early

            add_transaction_called.append(True)

            assert intrinsics["name"] == required_attrs["name"]

            # check that error event intrinsics haven't bled in

            assert "error.class" not in intrinsics
            assert "error.message" not in intrinsics
            assert "error.expected" not in intrinsics
            assert "transactionName" not in intrinsics

            _validate_event_attributes(
                intrinsics,
                user_attributes,
                required_attrs,
                required_user_attrs,
            )

            return wrapped(*args, **kwargs)

        result = _validate_transaction_event_sample_data(wrapped)(*args, **kwargs)
        assert add_transaction_called
        return result

    return _validate_wrapper


def validate_transaction_error_event_count(num_errors=1):
    """Validate that the correct number of error events are saved to StatsEngine
    after a transaction
    """

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_error_event_on_stats_engine(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            error_events = list(instance.error_events)
            assert len(error_events) == num_errors

        return result

    return _validate_error_event_on_stats_engine


def validate_transaction_error_trace_count(num_errors):
    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_transaction_error_trace_count(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            traced_errors = instance.error_data()
            assert len(traced_errors) == num_errors

        return result

    return _validate_transaction_error_trace_count


def validate_stats_engine_explain_plan_output_is_none():
    """This fixture isn't useful by itself, because you need to generate
    explain plans, which doesn't normally occur during record_transaction().

    Use the `validate_transaction_slow_sql_count` fixture to force the
    generation of slow sql data after record_transaction(), which will run
    newrelic.core.stats_engine.explain_plan.

    """

    @transient_function_wrapper("newrelic.core.stats_engine", "explain_plan")
    def _validate_explain_plan_output_is_none(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            assert result is None

        return result

    return _validate_explain_plan_output_is_none


def validate_error_event_sample_data(required_attrs=None, required_user_attrs=True, num_errors=1):
    """Validate the data collected for error_events. This test depends on values
    in the test application from agent_features/test_analytics.py, and is only
    meant to be run as a validation with those tests.
    """
    required_attrs = required_attrs or {}

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_error_event_sample_data(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:

            def _bind_params(transaction, *args, **kwargs):
                return transaction

            transaction = _bind_params(*args, **kwargs)

            error_events = transaction.error_events(instance.stats_table)
            assert len(error_events) == num_errors
            for sample in error_events:
                assert isinstance(sample, list)
                assert len(sample) == 3

                intrinsics, user_attributes, _ = sample

                # These intrinsics should always be present

                assert intrinsics["type"] == "TransactionError"
                assert intrinsics["transactionName"] == required_attrs["transactionName"]
                assert intrinsics["error.class"] == required_attrs["error.class"]
                assert intrinsics["error.message"].startswith(required_attrs["error.message"])
                assert intrinsics["error.expected"] == required_attrs["error.expected"]
                assert intrinsics["nr.transactionGuid"] is not None
                assert intrinsics["spanId"] is not None

                # check that transaction event intrinsics haven't bled in

                assert "name" not in intrinsics

                _validate_event_attributes(intrinsics, user_attributes, required_attrs, required_user_attrs)
                if required_user_attrs:
                    error_user_params = error_user_params_added()
                    for param, value in error_user_params.items():
                        assert user_attributes[param] == value

        return result

    return _validate_error_event_sample_data


def _validate_event_attributes(intrinsics, user_attributes, required_intrinsics, required_user):
    now = time.time()
    assert isinstance(intrinsics["timestamp"], int)
    assert intrinsics["timestamp"] <= 1000.0 * now
    assert intrinsics["duration"] >= 0.0

    assert "memcacheDuration" not in intrinsics

    if required_user:
        required_user_attributes = user_attributes_added()
        for attr, value in required_user_attributes.items():
            assert user_attributes[attr] == value
    else:
        assert user_attributes == {}

    if "databaseCallCount" in required_intrinsics:
        assert intrinsics["databaseDuration"] > 0
        call_count = required_intrinsics["databaseCallCount"]
        assert intrinsics["databaseCallCount"] == call_count
    else:
        assert "databaseDuration" not in intrinsics
        assert "databaseCallCount" not in intrinsics

    if "externalCallCount" in required_intrinsics:
        assert intrinsics["externalDuration"] > 0
        call_count = required_intrinsics["externalCallCount"]
        assert intrinsics["externalCallCount"] == call_count
    else:
        assert "externalDuration" not in intrinsics
        assert "externalCallCount" not in intrinsics

    if intrinsics.get("queueDuration", False):
        assert intrinsics["queueDuration"] > 0
    else:
        assert "queueDuration" not in intrinsics

    if "nr.referringTransactionGuid" in required_intrinsics:
        guid = required_intrinsics["nr.referringTransactionGuid"]
        assert intrinsics["nr.referringTransactionGuid"] == guid
    else:
        assert "nr.referringTransactionGuid" not in intrinsics

    if "nr.syntheticsResourceId" in required_intrinsics:
        res_id = required_intrinsics["nr.syntheticsResourceId"]
        job_id = required_intrinsics["nr.syntheticsJobId"]
        monitor_id = required_intrinsics["nr.syntheticsMonitorId"]
        assert intrinsics["nr.syntheticsResourceId"] == res_id
        assert intrinsics["nr.syntheticsJobId"] == job_id
        assert intrinsics["nr.syntheticsMonitorId"] == monitor_id

    if "port" in required_intrinsics:
        assert intrinsics["port"] == required_intrinsics["port"]


def validate_transaction_exception_message(expected_message):
    """Test exception message encoding/decoding for a single error"""

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_transaction_exception_message(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            error_data = instance.error_data()
            assert len(error_data) == 1
            error = error_data[0]

            # make sure we have the one error we are testing

            # Because we ultimately care what is sent to APM, run the exception
            # data through the encoding code that is would be run through
            # before being sent to the collector.

            encoded_error = json_encode(error)

            # to decode, use un-adultered json loading methods

            decoded_json = json.loads(encoded_error)

            message = decoded_json[2]
            assert expected_message == message

        return result

    return _validate_transaction_exception_message


def validate_application_exception_message(expected_message):
    """Test exception message encoding/decoding for a single error"""

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.notice_error")
    def _validate_application_exception_message(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            error_data = instance.error_data()
            assert len(error_data) == 1
            error = error_data[0]

            # make sure we have the one error we are testing

            # Because we ultimately care what is sent to APM, run the exception
            # data through the encoding code that is would be run through
            # before being sent to the collector.

            encoded_error = json_encode(error)

            # to decode, use un-adultered json loading methods

            decoded_json = json.loads(encoded_error)

            message = decoded_json[2]
            assert expected_message == message

        return result

    return _validate_application_exception_message


def _validate_custom_event(recorded_event, required_event):
    assert len(recorded_event) == 2  # [intrinsic, user attributes]

    intrinsics = recorded_event[0]

    assert intrinsics["type"] == required_event[0]["type"]

    now = time.time()
    assert isinstance(intrinsics["timestamp"], int)
    assert intrinsics["timestamp"] <= 1000.0 * now
    assert intrinsics["timestamp"] >= 1000.0 * required_event[0]["timestamp"]

    assert recorded_event[1].items() == required_event[1].items()


def validate_custom_event_in_application_stats_engine(required_event):
    @function_wrapper
    def _validate_custom_event_in_application_stats_engine(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            stats = core_application_stats_engine(None)
            assert stats.custom_events.num_samples == 1

            custom_event = next(iter(stats.custom_events))
            _validate_custom_event(custom_event, required_event)

        return result

    return _validate_custom_event_in_application_stats_engine


def validate_custom_event_count(count):
    @function_wrapper
    def _validate_custom_event_count(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            stats = core_application_stats_engine(None)
            assert stats.custom_events.num_samples == count

        return result

    return _validate_custom_event_count


def _validate_node_parenting(node, expected_node):
    assert node.exclusive >= 0, "node.exclusive = %s" % node.exclusive

    expected_children = expected_node[1]

    def len_error():
        return ("len(node.children)=%s, len(expected_children)=%s, node.children=%s") % (
            len(node.children),
            len(expected_children),
            node.children,
        )

    assert len(node.children) == len(expected_children), len_error()

    for index, child in enumerate(node.children):
        assert child.start_time > node.start_time
        _validate_node_parenting(child, expected_children[index])


def validate_tt_parenting(expected_parenting):
    """
    Validate the parenting and start_time of each node in a transaction trace

    expected_parenting is a tuple. The second item is a list of child nodes.
    The first item is not used in validation and exists as a tool for the
    developer to differentiate the node tuples.

        expected_parenting_example = (
            'TransactionNode', [
                ('FunctionNode', [
                    ('FunctionNode', [
                        ('FunctionNode', []),
                        ('FunctionNode', []),
                    ]),
            ]),
        ])
    """

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_tt_parenting(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        finally:

            def _bind_params(transaction, *args, **kwargs):
                return transaction

            transaction = _bind_params(*args, **kwargs)
            _validate_node_parenting(transaction.root, expected_parenting)

        return result

    return _validate_tt_parenting


def override_application_name(app_name):
    # The argument here cannot be named 'name', or else it triggers
    # a PyPy bug. Hence, we use 'app_name' instead.

    class Application(ObjectProxy):
        @property
        def name(self):
            return app_name

    @transient_function_wrapper("newrelic.api.transaction", "Transaction.__init__")
    def _override_application_name(wrapped, instance, args, kwargs):
        def _bind_params(application, *args, **kwargs):
            return application, args, kwargs

        application, _args, _kwargs = _bind_params(*args, **kwargs)

        application = Application(application)

        return wrapped(application, *_args, **_kwargs)

    return _override_application_name


@function_wrapper
def dt_enabled(wrapped, instance, args, kwargs):
    @transient_function_wrapper("newrelic.core.adaptive_sampler", "AdaptiveSampler.compute_sampled")
    def force_sampled(wrapped, instance, args, kwargs):
        wrapped(*args, **kwargs)
        return True

    settings = {"distributed_tracing.enabled": True}
    wrapped = override_application_settings(settings)(wrapped)
    wrapped = force_sampled(wrapped)

    return wrapped(*args, **kwargs)  # pylint: disable=E1102


@function_wrapper
def cat_enabled(wrapped, instance, args, kwargs):
    settings = {"cross_application_tracer.enabled": True, "distributed_tracing.enabled": False}
    wrapped = override_application_settings(settings)(wrapped)

    return wrapped(*args, **kwargs)


def override_application_settings(overrides):
    @function_wrapper
    def _override_application_settings(wrapped, instance, args, kwargs):
        # The settings object has references from a number of
        # different places. We have to create a copy, overlay
        # the temporary settings and then when done clear the
        # top level settings object and rebuild it when done.
        original_settings = application_settings()
        backup = copy.deepcopy(original_settings.__dict__)

        try:
            for name, value in overrides.items():
                apply_config_setting(original_settings, name, value)

            # should also update the attribute filter since it is affected
            # by application settings

            flat_settings = flatten_settings(original_settings)
            original_settings.attribute_filter = AttributeFilter(flat_settings)

            return wrapped(*args, **kwargs)
        finally:
            original_settings.__dict__.clear()
            original_settings.__dict__.update(backup)

    return _override_application_settings


def override_generic_settings(settings_object, overrides):
    @function_wrapper
    def _override_generic_settings(wrapped, instance, args, kwargs):
        # In some cases, a settings object may have references
        # from a number of different places. We have to create
        # a copy, overlay the temporary settings and then when
        # done, clear the top level settings object and rebuild
        # it when done.
        original = settings_object
        backup = copy.deepcopy(original.__dict__)

        try:
            for name, value in overrides.items():
                apply_config_setting(original, name, value)
            return wrapped(*args, **kwargs)
        finally:
            original.__dict__.clear()
            original.__dict__.update(backup)

    return _override_generic_settings


def override_ignore_status_codes(status_codes):
    @function_wrapper
    def _override_ignore_status_codes(wrapped, instance, args, kwargs):
        # Updates can be made to ignored status codes in server
        # side configs. Changes will be applied to application
        # settings so we first check there and if they don't
        # exist, we default to global settings

        application = application_instance()
        settings = application and application.settings

        if not settings:
            settings = global_settings()

        original = settings.error_collector.ignore_status_codes

        try:
            settings.error_collector.ignore_status_codes = status_codes
            return wrapped(*args, **kwargs)
        finally:
            settings.error_collector.ignore_status_codes = original

    return _override_ignore_status_codes


def override_expected_status_codes(status_codes):
    @function_wrapper
    def _override_expected_status_codes(wrapped, instance, args, kwargs):
        # Updates can be made to expected status codes in server
        # side configs. Changes will be applied to application
        # settings so we first check there and if they don't
        # exist, we default to global settings

        application = application_instance()
        settings = application and application.settings

        if not settings:
            settings = global_settings()

        original = settings.error_collector.expected_status_codes

        try:
            settings.error_collector.expected_status_codes = status_codes
            return wrapped(*args, **kwargs)
        finally:
            settings.error_collector.expected_status_codes = original

    return _override_expected_status_codes


def code_coverage_fixture(source=None):
    if source is None:
        source = ["newrelic"]

    github_actions = bool(os.environ.get("GITHUB_ACTIONS", None))
    tox_env_directory = os.environ.get("TOX_ENVDIR", None)

    if tox_env_directory:
        data_file = os.path.join(tox_env_directory, ".coverage")
        data_suffix = os.path.split(tox_env_directory)[-1]
        coverage_directory = os.path.join(tox_env_directory, "htmlcov")
        xml_report = os.path.join(tox_env_directory, "coverage.xml")
    else:
        data_file = ".coverage"
        data_suffix = None
        coverage_directory = "htmlcov"
        xml_report = "coverage.xml"

    @pytest.fixture(scope="session")
    def _code_coverage_fixture(request):
        if not source:
            yield None  # Required, generator based fixtures must yield 1 value or pytest will throw an exception.
            return

        from coverage import coverage

        cov = coverage(source=source, data_file=data_file, data_suffix=data_suffix, branch=True)
        cov.start()

        yield cov

        # At exit, stop coverage and save to data file
        cov.stop()
        cov.save()
        if not github_actions:
            # Run html and xml reports locally
            cov.html_report(directory=coverage_directory)
            cov.xml_report(outfile=xml_report)

    return _code_coverage_fixture


def reset_core_stats_engine():
    """Reset the StatsEngine and custom StatsEngine of the core application."""

    @function_wrapper
    def _reset_core_stats_engine(wrapped, instance, args, kwargs):
        api_application = application_instance()
        api_name = api_application.name
        core_application = api_application._agent.application(api_name)

        stats = core_application._stats_engine
        stats.reset_stats(stats.settings)

        custom_stats = core_application._stats_custom_engine
        custom_stats.reset_stats(custom_stats.settings)

        return wrapped(*args, **kwargs)

    return _reset_core_stats_engine


def core_application_stats_engine(app_name=None):
    """Return the StatsEngine object from the core application object.

    Useful when validating items added outside of a transaction, since
    monkey-patching StatsEngine.record_transaction() doesn't work in
    those situations.

    """

    api_application = application_instance(app_name)
    api_name = api_application.name
    core_application = api_application._agent.application(api_name)
    return core_application._stats_engine


def core_application_stats_engine_error(error_type, app_name=None):
    """Return a single error with the type of error_type, or None.

    In the core application StatsEngine, look in StatsEngine.error_data()
    and return the first error with the type of error_type. If none found,
    return None.

    Useful for verifying that application.notice_error() works, since
    the error is saved outside of a transaction. Must use a unique error
    type per test in a single test file, so that it returns the error you
    expect. (If you have 2 tests that record the same type of exception, then
    StatsEngine.error_data() will contain 2 errors with the same type, but
    this function will always return the first one it finds.)

    """

    stats = core_application_stats_engine(app_name)
    errors = stats.error_data()
    return next((e for e in errors if e.type == error_type), None)


def error_is_saved(error, app_name=None):
    """Return True, if an error of a particular type has already been saved.

    Before calling application.notice_error() in a test, it's good to
    check if that type of error has already been saved, so you know that
    there will only be a single example of a type of Error in error_data()
    when you verify that the exception was recorded correctly.

    Example usage:

        try:
            assert not error_is_saved(ErrorOne)
            raise ErrorOne('error one message')
        except ErrorOne:
            application_instance = application()
            application_instance.notice_error()

        my_error = core_application_stats_engine_error(_error_one_name)
        assert my_error.message == 'error one message'

    """

    error_name = callable_name(error)
    stats = core_application_stats_engine(app_name)
    errors = stats.error_data()
    return error_name in [e.type for e in errors if e.type == error_name]


def set_default_encoding(encoding):
    """Changes the default encoding of the global environment. Only works in
    Python 2, will cause an error in Python 3
    """

    # If using this with other decorators/fixtures that depend on the system
    # default encoding, this decorator must be on wrapped on top of them.

    @function_wrapper
    def _set_default_encoding(wrapped, instance, args, kwargs):
        # This technique of reloading the sys module is necessary because the
        # method is removed during initialization of Python. Doing this is
        # highly frowned upon, but it is the only way to test how our agent
        # behaves when different sys encodings are used. For more information,
        # see this Stack Overflow post: http://bit.ly/1xBNxRc

        six.moves.reload_module(sys)  # pylint: disable=E1101
        original_encoding = sys.getdefaultencoding()
        sys.setdefaultencoding(encoding)  # pylint: disable=E1101

        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        finally:
            sys.setdefaultencoding(original_encoding)  # pylint: disable=E1101

        return result

    return _set_default_encoding


def function_not_called(module, name):
    """Verify that a function is not called.

    Assert False, if it is.

    """

    called = []

    @transient_function_wrapper(module, name)
    def _function_not_called_(wrapped, instance, args, kwargs):
        called.append(True)
        return wrapped(*args, **kwargs)

    @function_wrapper
    def wrapper(wrapped, instance, args, kwargs):
        new_wrapper = _function_not_called_(wrapped)
        result = new_wrapper(*args, **kwargs)
        assert not called
        return result

    return wrapper


def validate_analytics_catmap_data(name, expected_attributes=(), non_expected_attributes=()):
    samples = []

    @transient_function_wrapper("newrelic.core.stats_engine", "SampledDataSet.add")
    def _capture_samples(wrapped, instance, args, kwargs):
        def _bind_params(sample, *args, **kwargs):
            return sample

        sample = _bind_params(*args, **kwargs)
        samples.append(sample)
        return wrapped(*args, **kwargs)

    @function_wrapper
    def _validate_analytics_sample_data(wrapped, instance, args, kwargs):
        _new_wrapped = _capture_samples(wrapped)

        result = _new_wrapped(*args, **kwargs)
        # Check type of s[0] because it returns an integer if s is a LogEventNode
        _samples = [s for s in samples if not isinstance(s[0], int) and s[0]["type"] == "Transaction"]
        assert _samples, "No Transaction events captured."
        for sample in _samples:
            assert isinstance(sample, list)
            assert len(sample) == 3

            intrinsics, _, _ = sample

            assert intrinsics["type"] == "Transaction"
            assert intrinsics["name"] == name
            assert intrinsics["timestamp"] >= 0.0
            assert intrinsics["duration"] >= 0.0

            for key, value in expected_attributes.items():
                assert intrinsics[key] == value

            for key in non_expected_attributes:
                assert intrinsics.get(key) is None

        return result

    return _validate_analytics_sample_data


def count_transactions(count_list):
    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _increment_count(wrapped, instance, args, kwargs):
        count_list.append(True)
        return wrapped(*args, **kwargs)

    return _increment_count


def failing_endpoint(endpoint, raises=RetryDataForRequest, call_number=1):
    called_list = []

    @transient_function_wrapper("newrelic.core.agent_protocol", "AgentProtocol.send")
    def send_request_wrapper(wrapped, instance, args, kwargs):
        def _bind_params(method, *args, **kwargs):
            return method

        method = _bind_params(*args, **kwargs)

        if method == endpoint:
            called_list.append(True)
            if len(called_list) == call_number:
                raise raises()

        return wrapped(*args, **kwargs)

    return send_request_wrapper


class Environ(object):
    """Context manager for setting environment variables temporarily."""

    def __init__(self, **kwargs):
        self._original_environ = os.environ
        self._environ_dict = kwargs

    def __enter__(self):
        for key, val in self._environ_dict.items():
            os.environ[key] = str(val)

    def __exit__(self, type, value, traceback):  # pylint: disable=redefined-builtin
        os.environ.clear()
        os.environ = self._original_environ


class TerminatingPopen(subprocess.Popen):
    """Context manager will terminate process when exiting, instead of waiting."""

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):  # pylint: disable=redefined-builtin,arguments-differ
        if self.stdout:
            self.stdout.close()
        if self.stderr:
            self.stderr.close()
        if self.stdin:
            self.stdin.close()

        self.terminate()


@pytest.fixture()
def newrelic_caplog(caplog):
    logger = logging.getLogger("newrelic")
    logger.propagate = True

    yield caplog



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import (
    function_wrapper,
    transient_function_wrapper,
)
from testing_support.fixtures import catch_background_exceptions


def validate_transaction_errors(
    errors=None, required_params=None, forgone_params=None, expected_errors=None
):
    errors = errors or []
    required_params = required_params or []
    forgone_params = forgone_params or []
    expected_errors = expected_errors or []
    captured_errors = []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    @catch_background_exceptions
    def _capture_transaction_errors(wrapped, instance, args, kwargs):
        def _bind_params(transaction, *args, **kwargs):
            return transaction

        transaction = _bind_params(*args, **kwargs)
        captured = transaction.errors

        captured_errors.append(captured)

        return wrapped(*args, **kwargs)

    @function_wrapper
    def _validate_transaction_errors(wrapped, instance, args, kwargs):
        _new_wrapped = _capture_transaction_errors(wrapped)
        output = _new_wrapped(*args, **kwargs)

        expected = sorted(errors)

        if captured_errors:
            captured = captured_errors[0]
        else:
            captured = []

        if errors and isinstance(errors[0], (tuple, list)):
            compare_to = sorted([(e.type, e.message) for e in captured])
        else:
            compare_to = sorted([e.type for e in captured])

        assert expected == compare_to, "expected=%r, captured=%r, errors=%r" % (expected, compare_to, captured)

        for e in captured:
            assert e.span_id
            for name, value in required_params:
                assert name in e.custom_params, "name=%r, params=%r" % (name, e.custom_params)
                assert e.custom_params[name] == value, "name=%r, value=%r, params=%r" % (
                    name,
                    value,
                    e.custom_params,
                )

            for name, value in forgone_params:
                assert name not in e.custom_params, "name=%r, params=%r" % (name, e.custom_params)

            if e.type in expected_errors:
                assert e.expected is True
            else:
                assert e.expected is False

        return output

    return _validate_transaction_errors



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random
import string

import pika
import six
from compat import basic_consume
from testing_support.db_settings import rabbitmq_settings
from testing_support.fixtures import (
    cat_enabled,
    override_application_settings,
)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from newrelic.api.background_task import background_task
from newrelic.api.transaction import current_transaction

DB_SETTINGS = rabbitmq_settings()[0]

ENCODING_KEY = "".join(random.choice(string.ascii_lowercase) for _ in range(40))
_override_settings = {
    "cross_process_id": "1#1",
    "encoding_key": ENCODING_KEY,
    "trusted_account_ids": [1],
    "browser_monitoring.enabled": False,
}


@background_task()
def do_basic_publish(channel, QUEUE, properties=None):
    channel.basic_publish(
        exchange="",
        routing_key=QUEUE,
        body="Testing CAT 123",
        properties=properties,
    )


_test_cat_basic_consume_scoped_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/Default", None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/Default", None),
]
_test_cat_basic_consume_rollup_metrics = list(_test_cat_basic_consume_scoped_metrics)
_test_cat_basic_consume_rollup_metrics.append(("ClientApplication/1#1/all", 1))
if six.PY3:
    _txn_name = "test_cat:test_basic_consume_cat_headers.<locals>.on_receive"
else:
    _txn_name = "test_cat:on_receive"


@validate_transaction_metrics(
    _txn_name,
    scoped_metrics=_test_cat_basic_consume_scoped_metrics,
    rollup_metrics=_test_cat_basic_consume_rollup_metrics,
    background_task=True,
    group="Message/RabbitMQ/Exchange/Default",
)
def do_basic_consume(channel):
    channel.start_consuming()


@cat_enabled
@override_application_settings(_override_settings)
def test_basic_consume_cat_headers():
    def on_receive(ch, method, properties, msg):
        headers = properties.headers
        assert headers
        assert "NewRelicID" not in headers
        assert "NewRelicTransaction" not in headers
        assert msg == b"Testing CAT 123"
        txn = current_transaction()
        assert txn.client_cross_process_id == "1#1"
        assert txn.client_account_id == 1
        assert txn.client_application_id == 1
        ch.stop_consuming()

    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()
        queue_name = "TESTCAT-%s" % os.getpid()
        channel.queue_declare(queue_name, durable=False)

        properties = pika.BasicProperties()
        properties.headers = {"Hello": "World"}

        try:
            basic_consume(channel, queue_name, on_receive, auto_ack=False)
            do_basic_publish(channel, queue_name, properties=properties)
            do_basic_consume(channel)

        finally:
            channel.queue_delete(queue_name)


_test_cat_basic_get_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/Default", None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/Default", 1),
    # Verify basic_get doesn't create a CAT metric
    ("ClientApplication/1#1/all", None),
]


@validate_transaction_metrics(
    "test_cat:do_basic_get",
    scoped_metrics=_test_cat_basic_get_metrics,
    rollup_metrics=_test_cat_basic_get_metrics,
    background_task=True,
)
@background_task()
def do_basic_get(channel, QUEUE):
    _, properties, msg = channel.basic_get(QUEUE)
    headers = properties.headers
    assert headers
    assert "NewRelicID" not in headers
    assert "NewRelicTransaction" not in headers
    assert msg == b"Testing CAT 123"
    txn = current_transaction()
    assert txn.client_cross_process_id is None
    assert txn.client_account_id is None
    assert txn.client_application_id is None


@override_application_settings(_override_settings)
def test_basic_get_no_cat_headers():
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()
        queue_name = "TESTCAT-%s" % os.getpid()
        channel.queue_declare(queue_name, durable=False)

        properties = pika.BasicProperties()
        properties.headers = {"Hello": "World"}

        try:
            do_basic_publish(channel, queue_name, properties=properties)
            do_basic_get(channel, queue_name)
        finally:
            channel.queue_delete(queue_name)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import tornado.gen
import tornado.httpclient
import tornado.httputil
import tornado.ioloop
import tornado.web
import tornado.websocket
from tornado.routing import PathMatches


def dummy(*args, **kwargs):
    pass


class BadGetStatusHandler(tornado.web.RequestHandler):
    def get(self):
        self.write("Hello, world")

    def get_status(self, *args, **kwargs):
        raise ValueError("Bad Status")


class ProcessCatHeadersHandler(tornado.web.RequestHandler):
    def __init__(self, application, request, response_code=200, **kwargs):
        super(ProcessCatHeadersHandler, self).__init__(application, request, **kwargs)
        self.response_code = response_code

    def get(self, client_cross_process_id, txn_header, flush=None):
        import newrelic.api.transaction as _transaction

        txn = _transaction.current_transaction()
        if txn:
            txn._process_incoming_cat_headers(client_cross_process_id, txn_header)

        if self.response_code != 200:
            self.set_status(self.response_code)
            return

        self.write("Hello, world")

        if flush == "flush":
            # Force a flush prior to calling finish
            # This causes the headers to get written immediately. The tests
            # which hit this endpoint will check that the response has been
            # properly processed even though we send the headers here.
            self.flush()

            # change the headers to garbage
            self.set_header("Content-Type", "garbage")


class EchoHeaderHandler(tornado.web.RequestHandler):
    def get(self):
        response = str(self.request.headers.__dict__).encode("utf-8")
        self.write(response)


class SimpleHandler(tornado.web.RequestHandler):
    options = {"your_command": "options"}

    def get(self):
        self.write("Hello, world")

    def log_exception(self, *args, **kwargs):
        pass

    post = get
    put = get
    delete = get
    patch = get


class NativeSimpleHandler(tornado.web.RequestHandler):
    async def get(self):
        self.write("Hello, world")


class SuperSimpleHandler(SimpleHandler):
    def get(self):
        super(SuperSimpleHandler, self).get()


class CallSimpleHandler(tornado.web.RequestHandler):
    def get(self):
        SimpleHandler(self.application, self.request).get()


class CoroThrowHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        try:
            yield self.throw_exception()
        except ValueError:
            pass
        self.write("Hello, world")

    @tornado.gen.coroutine
    def throw_exception(self):
        raise ValueError("Throwing exception.")


class CoroHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        yield tornado.gen.sleep(0)
        self.write("Hello, world")


class FakeCoroHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        self.write("Hello, world")


class InitializeHandler(tornado.web.RequestHandler):
    def initialize(self, *args, **kwargs):
        pass

    def get(self):
        self.write("Hello, world")


class HTMLInsertionHandler(tornado.web.RequestHandler):
    HTML = """<html>
    <head>
        <meta charset="utf-8">
        <title>My Website</title>
    </head>
    <body>
        Hello World! There is no New Relic Browser here :(
    </body>
    </html>
    """

    def get(self):
        self.finish(self.HTML)


class CrashHandler(tornado.web.RequestHandler):
    def get(self):
        raise ValueError("CrashHandler")


class MultiTraceHandler(tornado.web.RequestHandler):
    async def get(self):
        coros = (self.trace() for _ in range(2))
        await tornado.gen.multi(coros)
        self.write("*")

    def trace(self):
        from newrelic.api.function_trace import FunctionTrace

        with FunctionTrace(name="trace", terminal=True):
            pass


class WebSocketHandler(tornado.websocket.WebSocketHandler):
    def on_message(self, message):
        self.write_message("hello " + message)


class EnsureFutureHandler(tornado.web.RequestHandler):
    def get(self):
        import asyncio

        async def coro_trace():
            from newrelic.api.function_trace import FunctionTrace

            with FunctionTrace(name="trace", terminal=True):
                await tornado.gen.sleep(0)

        asyncio.ensure_future(coro_trace())


class WebNestedHandler(WebSocketHandler):
    def on_message(self, message):
        super(WebNestedHandler, self).on_message(message)


class CustomApplication(tornado.httputil.HTTPServerConnectionDelegate, tornado.httputil.HTTPMessageDelegate):
    def start_request(self, server_conn, http_conn):
        self.server_conn = server_conn
        self.http_conn = http_conn
        return self

    def finish(self):
        response_line = tornado.httputil.ResponseStartLine("HTTP/1.1", 200, "OK")
        headers = tornado.httputil.HTTPHeaders()
        headers["Content-Type"] = "text/plain"
        self.http_conn.write_headers(response_line, headers)
        self.http_conn.write(b"*")
        self.http_conn.finish()


class BlockingHandler(tornado.web.RequestHandler):
    total = 0
    future = None

    def initialize(self, yield_before_finish=False):
        self.yield_before_finish = yield_before_finish

    async def get(self, total=1):
        import asyncio

        total = int(total)

        cls = type(self)
        if cls.total == 0:
            cls.future = asyncio.Future()

        cls.total += 1
        if cls.total == total:
            cls.total = 0
            cls.future.set_result(True)
            cls.future = None
            time.sleep(0.1)
        else:
            await cls.future

        if self.yield_before_finish:
            await asyncio.sleep(0)

        self.write("*")


def make_app(custom=False):
    handlers = [
        (PathMatches(r"/simple"), SimpleHandler),
        (r"/crash", CrashHandler),
        (r"/call-simple", CallSimpleHandler),
        (r"/super-simple", SuperSimpleHandler),
        (r"/coro", CoroHandler),
        (r"/coro-throw", CoroThrowHandler),
        (r"/fake-coro", FakeCoroHandler),
        (r"/init", InitializeHandler),
        (r"/html-insertion", HTMLInsertionHandler),
        (r"/bad-get-status", BadGetStatusHandler),
        (r"/force-cat-response/(\S+)/(\S+)/(\S+)", ProcessCatHeadersHandler),
        (r"/304-cat-response/(\S+)/(\S+)", ProcessCatHeadersHandler, {"response_code": 304}),
        (r"/echo-headers", EchoHeaderHandler),
        (r"/native-simple", NativeSimpleHandler),
        (r"/multi-trace", MultiTraceHandler),
        (r"/web-socket", WebSocketHandler),
        (r"/ensure-future", EnsureFutureHandler),
        (r"/call-web-socket", WebNestedHandler),
        (r"/block/(\d+)", BlockingHandler),
        (r"/block-with-yield/(\d+)", BlockingHandler, {"yield_before_finish": True}),
    ]
    if custom:
        return CustomApplication()
    else:
        return tornado.web.Application(handlers, log_function=dummy)


if __name__ == "__main__":
    app = make_app()
    app.listen(8888, address="127.0.0.1")
    tornado.ioloop.IOLoop.current().start()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from testing_support.fixtures import catch_background_exceptions

from newrelic.common.object_wrapper import transient_function_wrapper

"""
operation: method name
target: search argument
"""


def validate_datastore_trace_inputs(operation=None, target=None):
    @transient_function_wrapper("newrelic.api.datastore_trace", "DatastoreTrace.__init__")
    @catch_background_exceptions
    def _validate_datastore_trace_inputs(wrapped, instance, args, kwargs):
        def _bind_params(product, target, operation, host=None, port_path_or_id=None, database_name=None, **kwargs):
            return (product, target, operation, host, port_path_or_id, database_name, kwargs)

        (
            captured_product,
            captured_target,
            captured_operation,
            captured_host,
            captured_port_path_or_id,
            captured_database_name,
            captured_kwargs,
        ) = _bind_params(*args, **kwargs)

        if target is not None:
            assert captured_target == target, "%s didn't match expected %s" % (captured_target, target)
        if operation is not None:
            assert captured_operation == operation, "%s didn't match expected %s" % (captured_operation, operation)

        return wrapped(*args, **kwargs)

    return _validate_datastore_trace_inputs



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is an implementation of the cross agent tests for cat map using a wsgi
application. Another implementation of these tests using a tornado application
can be found in test/framework_tornado_r3/test_cat_map.py
"""

import webtest
import pytest
import json
import os

try:
    from urllib2 import urlopen  # Py2.X
except ImportError:
    from urllib.request import urlopen   # Py3.X

from newrelic.packages import six

from newrelic.api.external_trace import ExternalTrace
from newrelic.api.transaction import (get_browser_timing_header,
        set_transaction_name, get_browser_timing_footer, set_background_task,
        current_transaction)
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.encoding_utils import obfuscate, json_encode

from testing_support.fixtures import (override_application_settings,
        override_application_name, validate_tt_parameters,
        make_cross_agent_headers, validate_analytics_catmap_data)
from testing_support.mock_external_http_server import (
        MockExternalHTTPHResponseHeadersServer)

ENCODING_KEY = '1234567890123456789012345678901234567890'
CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
JSON_DIR = os.path.normpath(os.path.join(CURRENT_DIR, 'fixtures'))
OUTBOUD_REQUESTS = {}

_parameters_list = ["name", "appName", "transactionName", "transactionGuid",
        "inboundPayload", "outboundRequests", "expectedIntrinsicFields",
        "nonExpectedIntrinsicFields"]


@pytest.fixture(scope='module')
def server():
    with MockExternalHTTPHResponseHeadersServer() as _server:
        yield _server


def load_tests():
    result = []
    path = os.path.join(JSON_DIR, 'cat_map.json')
    with open(path, 'r') as fh:
        tests = json.load(fh)

    for test in tests:
        values = tuple([test.get(param, None) for param in _parameters_list])
        result.append(values)

    return result


_parameters = ",".join(_parameters_list)


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status = '200 OK'

    txn_name = environ.get('txn')
    if six.PY2:
        txn_name = txn_name.decode('UTF-8')
    txn_name = txn_name.split('/', 3)

    guid = environ.get('guid')
    old_cat = environ.get('old_cat') == 'True'
    txn = current_transaction()

    txn.guid = guid
    for req in OUTBOUD_REQUESTS:
        # Change the transaction name before making an outbound call.
        outgoing_name = req['outboundTxnName'].split('/', 3)
        if outgoing_name[0] != 'WebTransaction':
            set_background_task(True)

        set_transaction_name(outgoing_name[2], group=outgoing_name[1])

        expected_outbound_header = obfuscate(
                json_encode(req['expectedOutboundPayload']), ENCODING_KEY)
        generated_outbound_header = dict(
                ExternalTrace.generate_request_headers(txn))

        # A 500 error is returned because 'assert' statements in the wsgi app
        # are ignored.

        if old_cat:
            if (expected_outbound_header !=
                    generated_outbound_header['X-NewRelic-Transaction']):
                status = '500 Outbound Headers Check Failed.'
        else:
            if 'X-NewRelic-Transaction' in generated_outbound_header:
                status = '500 Outbound Headers Check Failed.'
        r = urlopen(environ['server_url'])
        r.read(10)

    # Set the final transaction name.

    if txn_name[0] != 'WebTransaction':
        set_background_task(True)
    set_transaction_name(txn_name[2], group=txn_name[1])

    text = '<html><head>%s</head><body><p>RESPONSE</p>%s</body></html>'

    output = (text % (get_browser_timing_header(),
            get_browser_timing_footer())).encode('UTF-8')

    response_headers = [('Content-type', 'text/html; charset=utf-8'),
                        ('Content-Length', str(len(output)))]
    start_response(status, response_headers)

    return [output]


target_application = webtest.TestApp(target_wsgi_application)


@pytest.mark.parametrize(_parameters, load_tests())
@pytest.mark.parametrize('old_cat', (True, False))
def test_cat_map(name, appName, transactionName, transactionGuid,
        inboundPayload, outboundRequests, expectedIntrinsicFields,
        nonExpectedIntrinsicFields, old_cat, server):
    global OUTBOUD_REQUESTS
    OUTBOUD_REQUESTS = outboundRequests or {}

    _custom_settings = {
            'cross_process_id': '1#1',
            'encoding_key': ENCODING_KEY,
            'trusted_account_ids': [1],
            'cross_application_tracer.enabled': True,
            'distributed_tracing.enabled': not old_cat,
            'transaction_tracer.transaction_threshold': 0.0,
    }

    if expectedIntrinsicFields and old_cat:
        _external_node_params = {
                'path_hash': expectedIntrinsicFields['nr.pathHash'],
                'trip_id': expectedIntrinsicFields['nr.tripId'],
        }
    else:
        _external_node_params = []

    if not old_cat:
        # since no better cat headers will be generated, no intrinsics should
        # be added
        expectedIntrinsicFields = {}

    @validate_tt_parameters(required_params=_external_node_params)
    @validate_analytics_catmap_data(transactionName,
            expected_attributes=expectedIntrinsicFields,
            non_expected_attributes=nonExpectedIntrinsicFields)
    @override_application_settings(_custom_settings)
    @override_application_name(appName)
    def run_cat_test():

        if six.PY2:
            txn_name = transactionName.encode('UTF-8')
            guid = transactionGuid.encode('UTF-8')
        else:
            txn_name = transactionName
            guid = transactionGuid

        # Only generate old cat style headers. This will test to make sure we
        # are properly ignoring these headers when the agent is using better
        # cat.

        headers = make_cross_agent_headers(inboundPayload, ENCODING_KEY, '1#1')
        response = target_application.get('/', headers=headers,
                extra_environ={'txn': txn_name, 'guid': guid,
                    'old_cat': str(old_cat),
                    'server_url': 'http://localhost:%d' % server.port})

        # Validation of analytic data happens in the decorator.

        assert response.status == '200 OK'

        content = response.html.html.body.p.string

        # Validate actual body content as sanity check.

        assert content == 'RESPONSE'

    run_cat_test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import (
    override_generic_settings,
    override_ignore_status_codes,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.common.object_names import callable_name
from newrelic.core.config import global_settings
from newrelic.packages import six

TEST_APPLICATION_PREFIX = "_test_application.create_app.<locals>" if six.PY3 else "_test_application"


@pytest.fixture(params=["flask_restful", "flask_restx"])
def application(request):
    from _test_application import get_test_application

    if request.param == "flask_restful":
        import flask_restful as module
    elif request.param == "flask_restx":
        import flask_restx as module
    else:
        assert False

    if "propagate_exceptions" in request.fixturenames:
        propagate_exceptions = request.getfixturevalue("propagate_exceptions")
    else:
        propagate_exceptions = False
    return get_test_application(module, propagate_exceptions)


_test_application_index_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_application:index", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
]


@validate_code_level_metrics(TEST_APPLICATION_PREFIX + ".IndexResource", "get")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics("_test_application:index", scoped_metrics=_test_application_index_scoped_metrics)
def test_application_index(application):
    response = application.get("/index")
    response.mustcontain("hello")


_test_application_raises_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Function/_test_application:exception", 1),
]


@pytest.mark.parametrize(
    "exception,status_code,ignore_status_code,propagate_exceptions",
    [
        ("werkzeug.exceptions:HTTPException", 404, False, False),
        ("werkzeug.exceptions:HTTPException", 404, True, False),
        ("werkzeug.exceptions:HTTPException", 503, False, False),
        ("_test_application:CustomException", 500, False, False),
        ("_test_application:CustomException", 500, False, True),
    ],
)
def test_application_raises(exception, status_code, ignore_status_code, propagate_exceptions, application):
    @validate_code_level_metrics(TEST_APPLICATION_PREFIX + ".ExceptionResource", "get")
    @validate_transaction_metrics("_test_application:exception", scoped_metrics=_test_application_raises_scoped_metrics)
    def _test():
        try:
            application.get("/exception/%s/%i" % (exception, status_code), status=status_code, expect_errors=True)
        except Exception as e:
            assert propagate_exceptions

            # check that the exception is the expected one
            if callable_name(type(e)) != exception:
                raise

    if ignore_status_code:
        _test = validate_transaction_errors(errors=[])(_test)
        _test = override_ignore_status_codes([status_code])(_test)
    else:
        _test = validate_transaction_errors(errors=[exception])(_test)
        _test = override_ignore_status_codes([])(_test)

    _test()


def test_application_outside_transaction(application):

    _settings = global_settings()

    @override_generic_settings(_settings, {"enabled": False})
    def _test():
        application.get("/exception/werkzeug.exceptions:HTTPException/404", status=404)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.datastore_trace import wrap_datastore_trace

_pysolr_client_methods = ("search", "more_like_this", "suggest_terms", "add", "delete", "commit", "optimize", "extract")

_pysolr_admin_methods = ("status", "create", "reload", "rename", "swap", "unload", "load")


def instrument_pysolr(module):
    for name in _pysolr_client_methods:
        if hasattr(module.Solr, name):
            wrap_datastore_trace(module, "Solr.%s" % name, product="Solr", target=None, operation=name)

    if hasattr(module, "SolrCoreAdmin"):
        for name in _pysolr_admin_methods:
            if hasattr(module.SolrCoreAdmin, name):
                wrap_datastore_trace(
                    module, "SolrCoreAdmin.%s" % name, product="Solr", target=None, operation="admin.%s" % name
                )



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from testing_support.fixtures import (code_coverage_fixture,
        collector_agent_registration_fixture, collector_available_fixture)

_coverage_source = [
    'newrelic.hooks.framework_django',
    'newrelic.hooks.component_djangorestframework',
]

code_coverage = code_coverage_fixture(source=_coverage_source)

_default_settings = {
    'transaction_tracer.explain_threshold': 0.0,
    'transaction_tracer.transaction_threshold': 0.0,
    'transaction_tracer.stack_trace_threshold': 0.0,
    'debug.log_data_collector_payloads': True,
    'debug.record_transaction_failure': True,
    'debug.log_autorum_middleware': True,
}

collector_agent_registration = collector_agent_registration_fixture(
        app_name='Python Agent Test (component_djangorestframework)',
        default_settings=_default_settings)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from enum import Enum


class ResponseState(Enum):
    NOT_STARTED = 1
    BODY = 2
    DONE = 3


class Response(object):
    def __init__(self, status, headers, body):
        self.status = status
        self.headers = headers
        self.body = b"".join(body)


class AsgiTest(object):
    def __init__(self, asgi_application):
        self.asgi_application = asgi_application

    def make_request(self, method, path, params=None, headers=None, body=None):
        loop = asyncio.new_event_loop()
        coro = self.make_request_async(method, path, params, headers, body)
        return loop.run_until_complete(coro)

    def get(self, path, params=None, headers=None, body=None):
        return self.make_request("GET", path, params, headers, body)

    async def make_request_async(self, method, path, params, headers, body):
        self.input_queue = asyncio.Queue()
        self.output_queue = asyncio.Queue()
        self.response_state = ResponseState.NOT_STARTED

        scope = self.generate_input(method, path, params, headers, body)

        try:
            awaitable = self.asgi_application(scope, self.input_queue.get, self.output_queue.put)
        except TypeError:
            instance = self.asgi_application(scope)
            awaitable = instance(self.input_queue.get, self.output_queue.put)

        await awaitable
        return self.process_output()

    def generate_input(self, method, path, params, headers, body):
        headers_list = []
        if headers:
            try:
                iterable = headers.items()
            except AttributeError:
                iterable = iter(headers)

            for key, value in iterable:
                header_tuple = (key.lower().encode("utf-8"), value.encode("utf-8"))
                headers_list.append(header_tuple)
        if not params:
            path, _, params = path.partition("?")

        scope = {
            "asgi": {"spec_version": "2.1", "version": "3.0"},
            "client": ("127.0.0.1", 54768),
            "headers": headers_list,
            "http_version": "1.1",
            "method": method.upper(),
            "path": path,
            "query_string": params and params.encode("utf-8") or b"",
            "raw_path": path.encode("utf-8"),
            "root_path": "",
            "scheme": "http",
            "server": ("127.0.0.1", 8000),
            "type": "http",
        }
        message = {"type": "http.request"}
        if body:
            message["body"] = body.encode("utf-8")

        self.input_queue.put_nowait(message)
        return scope

    def process_output(self):
        response_status = None
        response_headers = []
        body = []

        while True:
            try:
                message = self.output_queue.get_nowait()
            except asyncio.QueueEmpty:
                break

            if self.response_state is ResponseState.NOT_STARTED:
                assert message["type"] == "http.response.start"
                response_status = message["status"]
                response_headers = message.get("headers", response_headers)
                self.response_state = ResponseState.BODY
            elif self.response_state is ResponseState.BODY:
                assert message["type"] == "http.response.body"
                body.append(message.get("body", b""))

                more_body = message.get("more_body", False)
                if not more_body:
                    self.response_state = ResponseState.DONE
            else:
                assert False, "ASGI protocol error: unexpected message"

        assert (
            self.response_state is ResponseState.DONE
        ), "ASGI protocol error: state is not DONE, expected additional messages"
        final_headers = {k.decode("utf-8"): v.decode("utf-8") for k, v in response_headers}
        assert len(final_headers) == len(response_headers), "Duplicate header names are not supported"
        return Response(response_status, final_headers, body)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import tempfile

import pytest
from testing_support.fixtures import (  # noqa: F401; pylint: disable=W0611
    collector_agent_registration_fixture,
    collector_available_fixture,
)
from testing_support.fixtures import (  # noqa: F401; pylint: disable=W0611
    newrelic_caplog as caplog,
)

from newrelic.core.agent import agent_instance

_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (agent_unittests)", default_settings=_default_settings
)


try:
    # python 2.x
    reload
except NameError:
    # python 3.x
    from importlib import reload


class FakeProtos(object):
    Span = object()
    SpanBatch = object()


sys.modules["grpc"] = object()
sys.modules["newrelic.core.infinite_tracing_pb2"] = FakeProtos


@pytest.fixture(scope="function")
def global_settings(request, monkeypatch):
    ini_contents = request.getfixturevalue("ini")

    monkeypatch.delenv("NEW_RELIC_HOST", raising=False)
    monkeypatch.delenv("NEW_RELIC_LICENSE_KEY", raising=False)

    if "env" in request.fixturenames:
        env = request.getfixturevalue("env")
        for k, v in env.items():
            monkeypatch.setenv(k, v)

    import newrelic.config as config  # pylint: disable=R0402
    import newrelic.core.config as core_config  # pylint: disable=R0402

    original = {}
    for attr in dir(core_config):
        original[attr] = getattr(core_config, attr)

    agent = agent_instance()
    original_agent_config = agent._config

    reload(core_config)
    reload(config)

    with tempfile.NamedTemporaryFile() as ini_file:
        ini_file.write(ini_contents)
        ini_file.seek(0)

        config.initialize(ini_file.name)

    agent._config = core_config.global_settings()

    yield core_config.global_settings

    monkeypatch.undo()
    for attr_name, attr_value in original.items():
        setattr(core_config, attr_name, attr_value)
    agent._config = original_agent_config
    reload(config)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import urllib3
import urllib3.connectionpool

try:
    import urllib3.connection
except ImportError:
    pass

from testing_support.external_fixtures import (
    cache_outgoing_headers,
    insert_incoming_headers,
)
from testing_support.fixtures import (
    cat_enabled,
    override_application_settings,
)
from testing_support.util import version2tuple
from testing_support.validators.validate_cross_process_headers import (
    validate_cross_process_headers,
)
from testing_support.validators.validate_external_node_params import (
    validate_external_node_params,
)
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from newrelic.api.background_task import background_task


@pytest.fixture(scope="session")
def metrics(server):
    scoped = [("External/localhost:%d/urllib3/" % server.port, 1)]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("External/localhost:%d/urllib3/" % server.port, 1),
    ]

    return scoped, rollup


def test_http_request_connection_pool_urlopen(server, metrics):
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_urllib3:test_http_request_connection_pool_urlopen",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_urllib3:test_http_request_connection_pool_urlopen")
    def _test():
        pool = urllib3.HTTPConnectionPool("localhost:%d" % server.port)
        pool.urlopen("GET", "/")

    _test()


def test_http_request_connection_pool_request(server, metrics):
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_urllib3:test_http_request_connection_pool_request",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_urllib3:test_http_request_connection_pool_request")
    def _test():
        pool = urllib3.HTTPConnectionPool("localhost:%d" % server.port)
        pool.request("GET", "/")

    _test()


def test_http_request_connection_from_url_request(server, metrics):
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_urllib3:test_http_request_connection_from_url_request",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_urllib3:test_http_request_connection_from_url_request")
    def _test():
        conn = urllib3.connection_from_url("http://localhost:%d" % server.port)
        conn.request("GET", "/")

    _test()


def test_http_request_pool_manager_urlopen(server, metrics):
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_urllib3:test_http_request_pool_manager_urlopen",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_urllib3:test_http_request_pool_manager_urlopen")
    def _test():
        pool = urllib3.PoolManager(5)
        pool.urlopen("GET", "http://localhost:%d/" % server.port)

    _test()


def test_https_request_connection_pool_urlopen(server, metrics):
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_urllib3:test_https_request_connection_pool_urlopen",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_urllib3:test_https_request_connection_pool_urlopen")
    def _test():
        # Setting retries to 0 so that metrics are recorded only once
        pool = urllib3.HTTPSConnectionPool("localhost:%d" % server.port, retries=0)
        try:
            pool.urlopen("GET", "/")
        except Exception:
            pass

    _test()


def test_https_request_connection_pool_request(server, metrics):
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_urllib3:test_https_request_connection_pool_request",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_urllib3:test_https_request_connection_pool_request")
    def _test():
        # Setting retries to 0 so that metrics are recorded only once
        pool = urllib3.HTTPSConnectionPool("localhost:%d" % server.port, retries=0)
        try:
            pool.request("GET", "/")
        except Exception:
            pass

    _test()


def test_port_included(server):
    scoped = [("External/localhost:%d/urllib3/" % server.port, 1)]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("External/localhost:%d/urllib3/" % server.port, 1),
    ]

    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_urllib3:test_port_included", scoped_metrics=scoped, rollup_metrics=rollup, background_task=True
    )
    @background_task(name="test_urllib3:test_port_included")
    def _test():
        conn = urllib3.connection_from_url("http://localhost:%d" % server.port)
        conn.request("GET", "/")

    _test()


# Starting in urllib3 1.8, urllib3 wrote their own version of the
# HTTPConnection class. Previously the httplib/http.client HTTPConnection class
# was used. We test httplib in a different test directory so we skip this test.
@pytest.mark.skipif(
    version2tuple(urllib3.__version__) < (1, 8), reason="urllib3.connection.HTTPConnection added in 1.8"
)
def test_HTTPConnection_port_included(server):
    scoped = [("External/localhost:%d/urllib3/" % server.port, 1)]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("External/localhost:%d/urllib3/" % server.port, 1),
    ]

    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_urllib3:test_HTTPConnection_port_included",
        scoped_metrics=scoped,
        rollup_metrics=rollup,
        background_task=True,
    )
    @background_task(name="test_urllib3:test_HTTPConnection_port_included")
    def _test():
        conn = urllib3.connection.HTTPConnection("localhost:%d" % server.port)
        conn.request("GET", "/")

    _test()


@pytest.mark.parametrize(
    "distributed_tracing,span_events",
    (
        (True, True),
        (True, False),
        (False, False),
    ),
)
def test_urlopen_cross_process_request(distributed_tracing, span_events, server):
    @validate_transaction_errors(errors=[])
    @background_task(name="test_urllib3:test_urlopen_cross_process_request")
    @cache_outgoing_headers
    @validate_cross_process_headers
    def _test():
        pool = urllib3.HTTPConnectionPool("localhost:%d" % server.port)
        pool.urlopen("GET", "/")

    _test = override_application_settings(
        {
            "distributed_tracing.enabled": distributed_tracing,
            "cross_application_tracer.enabled": not distributed_tracing,
            "span_events.enabled": span_events,
        }
    )(_test)

    _test()


@cat_enabled
def test_urlopen_cross_process_response(server):
    _test_urlopen_cross_process_response_scoped_metrics = [
        ("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1)
    ]

    _test_urlopen_cross_process_response_rollup_metrics = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("ExternalApp/localhost:%d/1#2/all" % server.port, 1),
        ("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1),
    ]

    _test_urlopen_cross_process_response_external_node_params = [
        ("cross_process_id", "1#2"),
        ("external_txn_name", "test"),
        ("transaction_guid", "0123456789012345"),
    ]

    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        "test_urllib3:test_urlopen_cross_process_response",
        scoped_metrics=_test_urlopen_cross_process_response_scoped_metrics,
        rollup_metrics=_test_urlopen_cross_process_response_rollup_metrics,
        background_task=True,
    )
    @insert_incoming_headers
    @validate_external_node_params(params=_test_urlopen_cross_process_response_external_node_params)
    @background_task(name="test_urllib3:test_urlopen_cross_process_response")
    def _test():
        pool = urllib3.HTTPConnectionPool("localhost:%d" % server.port)
        pool.urlopen("GET", "/")

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.http import HttpResponse
from django.views.generic.base import View, TemplateView
from django.shortcuts import render
from django.core.exceptions import PermissionDenied
from middleware import Custom410

from newrelic.api.transaction import (get_browser_timing_header,
    get_browser_timing_footer)


def index(request):
    return HttpResponse('INDEX RESPONSE')


def exception(request):
    raise RuntimeError('exception')


def permission_denied(request):
    raise PermissionDenied()


def middleware_410(request):
    raise Custom410()


class MyView(View):
    def get(self, request):
        return HttpResponse('CBV RESPONSE')


def deferred_cbv(request):
    return MyView.as_view()(request)


def html_insertion(request):
    return HttpResponse('<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')


def html_insertion_content_length(request):
    content = ('<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')
    response = HttpResponse(content)
    response['Content-Length'] = len(content)
    return response


def html_insertion_manual(request):
    header = get_browser_timing_header()
    footer = get_browser_timing_footer()

    header = get_browser_timing_header()
    footer = get_browser_timing_footer()

    assert header == ''
    assert footer == ''

    return HttpResponse('<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')


def html_insertion_unnamed_attachment_header(request):
    response = HttpResponse('<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')
    response['Content-Disposition'] = 'attachment'
    return response


def html_insertion_named_attachment_header(request):
    response = HttpResponse('<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')
    response['Content-Disposition'] = 'Attachment; filename="X"'
    return response


def inclusion_tag(request):
    return render(request, 'main.html', {}, content_type="text/html")


def template_tags(request):
    return render(request, 'main.html', {}, content_type="text/html")


def render_exception_function(request):
    return render(request, 'render_exception.html')


class RenderExceptionClass(TemplateView):
    template_name = 'render_exception.html'


def gzip_html_insertion(request):
    # contents must be at least 200 bytes for gzip middleware to work
    contents = '*' * 200
    return HttpResponse('<!DOCTYPE html><html><head>Some header</head>'
        '<body><h1>My First Heading</h1><p>%s</p></body></html>' % contents)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.encoding_utils import unpack_field
from newrelic.common.object_wrapper import transient_function_wrapper
from newrelic.core.database_utils import SQLConnections


def validate_synthetics_transaction_trace(required_params=None, forgone_params=None, should_exist=True):
    required_params = required_params or {}
    forgone_params = forgone_params or {}

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_synthetics_transaction_trace(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:

            # Now that transaction has been recorded, generate
            # a transaction trace

            connections = SQLConnections()
            trace_data = instance.transaction_trace_data(connections)

            # Check that synthetics resource id is in TT header

            header = trace_data[0]
            header_key = "synthetics_resource_id"

            if should_exist:
                assert header_key in required_params
                assert header[9] == required_params[header_key], "name=%r, header=%r" % (header_key, header)
            else:
                assert header[9] is None

            # Check that synthetics ids are in TT custom params

            pack_data = unpack_field(trace_data[0][4])
            tt_intrinsics = pack_data[0][4]["intrinsics"]

            for name in required_params:
                assert name in tt_intrinsics, "name=%r, intrinsics=%r" % (name, tt_intrinsics)
                assert tt_intrinsics[name] == required_params[name], "name=%r, value=%r, intrinsics=%r" % (
                    name,
                    required_params[name],
                    tt_intrinsics,
                )

            for name in forgone_params:
                assert name not in tt_intrinsics, "name=%r, intrinsics=%r" % (name, tt_intrinsics)

        return result

    return _validate_synthetics_transaction_trace



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import pytest
from testing_support.fixtures import (
    make_cross_agent_headers,
    override_application_settings,
    validate_analytics_catmap_data,
)
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)

from newrelic.common.encoding_utils import deobfuscate
from newrelic.common.object_wrapper import transient_function_wrapper

ENCODING_KEY = "1234567890123456789012345678901234567890"
test_uris = [
    ("/error?hello=world", "_target_application:error"),
    ("/coro?hello=world", "_target_application:index"),
    ("/class?hello=world", "_target_application:HelloWorldView._respond"),
]


def record_aiohttp1_raw_headers(raw_headers):
    try:
        import aiohttp.protocol  # noqa: F401
    except ImportError:

        def pass_through(function):
            return function

        return pass_through

    @transient_function_wrapper("aiohttp.protocol", "HttpParser.parse_headers")
    def recorder(wrapped, instance, args, kwargs):
        def _bind_params(lines):
            return lines

        lines = _bind_params(*args, **kwargs)
        for line in lines:
            line = line.decode("utf-8")

            # This is the request, not the response
            if line.startswith("GET"):
                break

            if ":" in line:
                key, value = line.split(":", maxsplit=1)
                raw_headers[key.strip()] = value.strip()

        return wrapped(*args, **kwargs)

    return recorder


@pytest.mark.parametrize(
    "inbound_payload,expected_intrinsics,forgone_intrinsics,cat_id",
    [
        # Valid payload from trusted account
        (
            ["b854df4feb2b1f06", False, "7e249074f277923d", "5d2957be"],
            {
                "nr.referringTransactionGuid": "b854df4feb2b1f06",
                "nr.tripId": "7e249074f277923d",
                "nr.referringPathHash": "5d2957be",
            },
            [],
            "1#1",
        ),
        # Valid payload from an untrusted account
        (
            ["b854df4feb2b1f06", False, "7e249074f277923d", "5d2957be"],
            {},
            ["nr.referringTransactionGuid", "nr.tripId", "nr.referringPathHash"],
            "80#1",
        ),
    ],
)
@pytest.mark.parametrize("method", ["GET"])
@pytest.mark.parametrize("uri,metric_name", test_uris)
def test_cat_headers(
    method, uri, metric_name, inbound_payload, expected_intrinsics, forgone_intrinsics, cat_id, aiohttp_app
):

    _raw_headers = {}

    async def fetch():
        headers = make_cross_agent_headers(inbound_payload, ENCODING_KEY, cat_id)
        resp = await aiohttp_app.client.request(method, uri, headers=headers)

        if _raw_headers:
            raw_headers = _raw_headers
        else:
            raw_headers = {k.decode("utf-8"): v.decode("utf-8") for k, v in resp.raw_headers}

        if expected_intrinsics:
            # test valid CAT response header
            assert "X-NewRelic-App-Data" in raw_headers

            app_data = json.loads(deobfuscate(raw_headers["X-NewRelic-App-Data"], ENCODING_KEY))
            assert app_data[0] == cat_id
            assert app_data[1] == ("WebTransaction/Function/%s" % metric_name)
        else:
            assert "X-NewRelic-App-Data" not in resp.headers

    _custom_settings = {
        "cross_process_id": "1#1",
        "encoding_key": ENCODING_KEY,
        "trusted_account_ids": [1],
        "cross_application_tracer.enabled": True,
        "distributed_tracing.enabled": False,
    }

    # NOTE: the logic-flow of this test can be a bit confusing.
    #       the override settings and attribute validation occur
    #       not when the request is made (above) since it does not
    #       occur inside a transaction. instead, the settings and
    #       validation are for the new transaction that is made
    #       asynchronously on the *server side* when the request
    #       is received and subsequently processed. that code is
    #       a fixture from conftest.py/_target_application.py

    @validate_analytics_catmap_data(
        "WebTransaction/Function/%s" % metric_name,
        expected_attributes=expected_intrinsics,
        non_expected_attributes=forgone_intrinsics,
    )
    @override_application_settings(_custom_settings)
    @record_aiohttp1_raw_headers(_raw_headers)
    def _test():
        aiohttp_app.loop.run_until_complete(fetch())

    _test()


account_id = "33"
primary_application_id = "2827902"

inbound_payload = {
    "v": [0, 1],
    "d": {
        "ac": account_id,
        "ap": primary_application_id,
        "id": "7d3efb1b173fecfa",
        "tx": "e8b91a159289ff74",
        "pr": 1.234567,
        "sa": True,
        "ti": 1518469636035,
        "tr": "d6b4ba0c3a712ca",
        "ty": "App",
    },
}

expected_attributes = {
    "agent": [],
    "user": [],
    "intrinsic": {
        "traceId": "d6b4ba0c3a712ca",
        "priority": 1.234567,
        "sampled": True,
        "parent.type": "App",
        "parent.app": primary_application_id,
        "parent.account": account_id,
        "parent.transportType": "HTTP",
        "parentId": "e8b91a159289ff74",
        "parentSpanId": "7d3efb1b173fecfa",
    },
}

unexpected_attributes = {
    "agent": [],
    "user": [],
    "intrinsic": ["grandparentId", "cross_process_id", "nr.tripId", "nr.pathHash"],
}


@pytest.mark.parametrize("uri,metric_name", test_uris)
def test_distributed_tracing_headers(uri, metric_name, aiohttp_app):
    async def fetch():
        headers = {"newrelic": json.dumps(inbound_payload)}
        resp = await aiohttp_app.client.request("GET", uri, headers=headers)

        # better cat does not send a response in the headers
        assert "newrelic" not in resp.headers

        # old-cat headers should not be in the response
        assert "X-NewRelic-App-Data" not in resp.headers

    # NOTE: the logic-flow of this test can be a bit confusing.
    #       the override settings and attribute validation occur
    #       not when the request is made (above) since it does not
    #       occur inside a transaction. instead, the settings and
    #       validation are for the new transaction that is made
    #       asynchronously on the *server side* when the request
    #       is received and subsequently processed. that code is
    #       a fixture from conftest.py/_target_application.py

    @validate_transaction_event_attributes(expected_attributes, unexpected_attributes)
    @override_application_settings(
        {
            "account_id": "33",
            "trusted_account_key": "33",
            "primary_application_id": primary_application_id,
            "distributed_tracing.enabled": True,
        }
    )
    def _test():
        aiohttp_app.loop.run_until_complete(fetch())

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import webtest
from testing_support.fixtures import (
    override_application_settings,
    validate_transaction_event_sample_data,
)
from testing_support.sample_applications import (
    fully_featured_app,
    user_attributes_added,
)
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)

from newrelic.api.application import application_settings
from newrelic.api.background_task import background_task
from newrelic.common.encoding_utils import deobfuscate
from newrelic.common.object_wrapper import transient_function_wrapper

fully_featured_application = webtest.TestApp(fully_featured_app)
_user_attributes = user_attributes_added()

# ====================== Test cases ====================================

_test_capture_attributes_enabled_settings = {"browser_monitoring.attributes.enabled": True}

_intrinsic_attributes = {
    "name": "WebTransaction/Uri/",
    "port": 80,
}


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=_user_attributes)
@override_application_settings(_test_capture_attributes_enabled_settings)
def test_capture_attributes_enabled():
    settings = application_settings()

    assert settings.browser_monitoring.enabled
    assert settings.browser_monitoring.attributes.enabled

    assert settings.js_agent_loader

    response = fully_featured_application.get("/")

    header = response.html.html.head.script.string
    content = response.html.html.body.p.string
    footer = response.html.html.body.script.string

    # Validate actual body content.

    assert content == "RESPONSE"

    # We no longer are in control of the JS contents of the header so
    # just check to make sure it contains at least the magic string
    # 'NREUM'.

    assert header.find("NREUM") != -1

    # Now validate the various fields of the footer related to analytics.
    # The fields are held by a JSON dictionary.

    data = json.loads(footer.split("NREUM.info=")[1])

    obfuscation_key = settings.license_key[:13]

    attributes = json.loads(deobfuscate(data["atts"], obfuscation_key))
    user_attrs = attributes["u"]

    # When you round-trip through json encoding and json decoding, you
    # always end up with unicode (unicode in Python 2, str in Python 3.)
    #
    # Previously, we would drop attribute values of type 'bytes' in Python 3.
    # Now, we accept them and `json_encode` uses an encoding of 'latin-1',
    # just like it does for Python 2. This only applies to attributes in browser
    # monitoring

    browser_attributes = _user_attributes.copy()

    browser_attributes["bytes"] = "bytes-value"
    browser_attributes["invalid-utf8"] = _user_attributes["invalid-utf8"].decode("latin-1")
    browser_attributes["multibyte-utf8"] = _user_attributes["multibyte-utf8"].decode("latin-1")

    for attr, value in browser_attributes.items():
        assert user_attrs[attr] == value, "attribute %r expected %r, found %r" % (attr, value, user_attrs[attr])


_test_no_attributes_recorded_settings = {"browser_monitoring.attributes.enabled": True}


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs={})
@override_application_settings(_test_no_attributes_recorded_settings)
def test_no_attributes_recorded():
    settings = application_settings()

    assert settings.browser_monitoring.enabled
    assert settings.browser_monitoring.attributes.enabled

    assert settings.js_agent_loader

    response = fully_featured_application.get("/", extra_environ={"record_attributes": "FALSE"})

    header = response.html.html.head.script.string
    content = response.html.html.body.p.string
    footer = response.html.html.body.script.string

    # Validate actual body content.

    assert content == "RESPONSE"

    # We no longer are in control of the JS contents of the header so
    # just check to make sure it contains at least the magic string
    # 'NREUM'.

    assert header.find("NREUM") != -1

    # Now validate the various fields of the footer related to analytics.
    # The fields are held by a JSON dictionary.

    data = json.loads(footer.split("NREUM.info=")[1])

    # As we are not recording any user or agent attributes, we should not
    # actually have an entry at all in the footer.

    assert "atts" not in data


_test_analytic_events_capture_attributes_disabled_settings = {
    "transaction_events.attributes.enabled": False,
    "browser_monitoring.attributes.enabled": True,
}


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs={})
@override_application_settings(_test_analytic_events_capture_attributes_disabled_settings)
def test_analytic_events_capture_attributes_disabled():
    settings = application_settings()

    assert settings.collect_analytics_events
    assert settings.transaction_events.enabled
    assert not settings.transaction_events.attributes.enabled

    assert settings.browser_monitoring.enabled
    assert settings.browser_monitoring.attributes.enabled

    assert settings.js_agent_loader

    response = fully_featured_application.get("/")

    header = response.html.html.head.script.string
    content = response.html.html.body.p.string
    footer = response.html.html.body.script.string

    # Validate actual body content.

    assert content == "RESPONSE"

    # We no longer are in control of the JS contents of the header so
    # just check to make sure it contains at least the magic string
    # 'NREUM'.

    assert header.find("NREUM") != -1

    # Now validate that attributes are present, since browser monitoring should
    # be enabled.

    data = json.loads(footer.split("NREUM.info=")[1])

    assert "atts" in data


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=_user_attributes)
def test_capture_attributes_default():
    settings = application_settings()

    assert settings.browser_monitoring.enabled
    assert not settings.browser_monitoring.attributes.enabled

    assert settings.js_agent_loader

    response = fully_featured_application.get("/")

    header = response.html.html.head.script.string
    content = response.html.html.body.p.string
    footer = response.html.html.body.script.string

    # Validate actual body content.

    assert content == "RESPONSE"

    # We no longer are in control of the JS contents of the header so
    # just check to make sure it contains at least the magic string
    # 'NREUM'.

    assert header.find("NREUM") != -1

    # Now validate that attributes are not present, since should
    # be disabled.

    data = json.loads(footer.split("NREUM.info=")[1])

    assert "atts" not in data


_test_analytic_events_background_task_settings = {"browser_monitoring.attributes.enabled": True}

_intrinsic_attributes = {"name": "OtherTransaction/Uri/"}


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=_user_attributes)
@override_application_settings(_test_analytic_events_background_task_settings)
def test_analytic_events_background_task():
    settings = application_settings()

    assert settings.collect_analytics_events
    assert settings.transaction_events.enabled

    assert settings.browser_monitoring.enabled
    assert settings.browser_monitoring.attributes.enabled

    assert settings.js_agent_loader

    response = fully_featured_application.get("/", extra_environ={"newrelic.set_background_task": True})

    assert response.html.html.head.script is None


_test_capture_attributes_disabled_settings = {"browser_monitoring.attributes.enabled": False}

_intrinsic_attributes = {"name": "WebTransaction/Uri/"}


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=_user_attributes)
@override_application_settings(_test_capture_attributes_disabled_settings)
def test_capture_attributes_disabled():
    settings = application_settings()

    assert settings.browser_monitoring.enabled
    assert not settings.browser_monitoring.attributes.enabled

    assert settings.js_agent_loader

    response = fully_featured_application.get("/")

    header = response.html.html.head.script.string
    content = response.html.html.body.p.string
    footer = response.html.html.body.script.string

    # Validate actual body content.

    assert content == "RESPONSE"

    # We no longer are in control of the JS contents of the header so
    # just check to make sure it contains at least the magic string
    # 'NREUM'.

    assert header.find("NREUM") != -1

    # Now validate that attributes are not present, since should
    # be disabled.

    data = json.loads(footer.split("NREUM.info=")[1])

    assert "atts" not in data


@transient_function_wrapper("newrelic.core.stats_engine", "SampledDataSet.add")
def validate_no_analytics_sample_data(wrapped, instance, args, kwargs):
    assert False, "Should not be recording analytic event."
    return wrapped(*args, **kwargs)


_test_collect_analytic_events_disabled_settings = {
    "collect_analytics_events": False,
    "distributed_tracing.enabled": False,
    "browser_monitoring.attributes.enabled": True,
}


@validate_no_analytics_sample_data
@override_application_settings(_test_collect_analytic_events_disabled_settings)
def test_collect_analytic_events_disabled():
    settings = application_settings()

    assert not settings.collect_analytics_events

    assert settings.browser_monitoring.enabled
    assert settings.browser_monitoring.attributes.enabled

    assert settings.js_agent_loader

    response = fully_featured_application.get("/")

    header = response.html.html.head.script.string
    content = response.html.html.body.p.string
    footer = response.html.html.body.script.string

    # Validate actual body content.

    assert content == "RESPONSE"

    # We no longer are in control of the JS contents of the header so
    # just check to make sure it contains at least the magic string
    # 'NREUM'.

    assert header.find("NREUM") != -1

    # Now validate that attributes are present, since should
    # be enabled.

    data = json.loads(footer.split("NREUM.info=")[1])

    assert "atts" in data


_test_analytic_events_disabled_settings = {
    "transaction_events.enabled": False,
    "distributed_tracing.enabled": False,
    "browser_monitoring.attributes.enabled": True,
}


@validate_no_analytics_sample_data
@override_application_settings(_test_analytic_events_disabled_settings)
def test_analytic_events_disabled():
    settings = application_settings()

    assert settings.collect_analytics_events
    assert not settings.transaction_events.enabled

    assert settings.browser_monitoring.enabled
    assert settings.browser_monitoring.attributes.enabled

    assert settings.js_agent_loader

    response = fully_featured_application.get("/")

    header = response.html.html.head.script.string
    content = response.html.html.body.p.string
    footer = response.html.html.body.script.string

    # Validate actual body content.

    assert content == "RESPONSE"

    # We no longer are in control of the JS contents of the header so
    # just check to make sure it contains at least the magic string
    # 'NREUM'.

    assert header.find("NREUM") != -1

    # Now validate that attributes are present, since should
    # be enabled.

    data = json.loads(footer.split("NREUM.info=")[1])

    assert "atts" in data


# -------------- Test call counts in analytic events ----------------


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=_user_attributes)
def test_no_database_or_external_attributes_in_analytics():
    """Make no external calls or database calls in the transaction and check
    if the analytic event doesn't have the databaseCallCount, databaseDuration,
    externalCallCount and externalDuration attributes.

    """
    settings = application_settings()

    assert settings.browser_monitoring.enabled

    response = fully_featured_application.get("/")

    # Validation of analytic data happens in the decorator.

    content = response.html.html.body.p.string

    # Validate actual body content.

    assert content == "RESPONSE"


_intrinsic_attributes = {
    "name": "WebTransaction/Uri/db",
    "databaseCallCount": 2,
}


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=_user_attributes)
def test_database_attributes_in_analytics():
    """Make database calls in the transaction and check if the analytic
    event has the databaseCallCount and databaseDuration attributes.

    """
    settings = application_settings()

    assert settings.browser_monitoring.enabled

    test_environ = {
        "db": "2",
    }
    response = fully_featured_application.get("/db", extra_environ=test_environ)

    # Validation of analytic data happens in the decorator.

    content = response.html.html.body.p.string

    # Validate actual body content.

    assert content == "RESPONSE"


_intrinsic_attributes = {
    "name": "WebTransaction/Uri/ext",
    "externalCallCount": 2,
}


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=_user_attributes)
def test_external_attributes_in_analytics():
    """Make external calls in the transaction and check if the analytic
    event has the externalCallCount and externalDuration attributes.

    """
    settings = application_settings()

    assert settings.browser_monitoring.enabled

    test_environ = {
        "external": "2",
    }
    response = fully_featured_application.get("/ext", extra_environ=test_environ)

    # Validation of analytic data happens in the decorator.

    content = response.html.html.body.p.string

    # Validate actual body content.

    assert content == "RESPONSE"


_intrinsic_attributes = {
    "name": "WebTransaction/Uri/dbext",
    "databaseCallCount": 2,
    "externalCallCount": 2,
}


@validate_transaction_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=_user_attributes)
def test_database_and_external_attributes_in_analytics():
    """Make external calls and database calls in the transaction and check if
    the analytic event has the databaseCallCount, databaseDuration,
    externalCallCount and externalDuration attributes.

    """
    settings = application_settings()

    assert settings.browser_monitoring.enabled

    test_environ = {
        "db": "2",
        "external": "2",
    }
    response = fully_featured_application.get("/dbext", extra_environ=test_environ)

    # Validation of analytic data happens in the decorator.

    content = response.html.html.body.p.string

    # Validate actual body content.

    assert content == "RESPONSE"


# -------------- Test background tasks ----------------

_expected_attributes = {
    "user": [],
    "agent": [],
    "intrinsic": ("name", "duration", "type", "timestamp", "totalTime"),
}

_expected_absent_attributes = {
    "user": ("foo"),
    "agent": ("response.status", "request.method"),
    "intrinsic": ("port"),
}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@background_task()
def test_background_task_intrinsics_has_no_port():
    pass
